# -*- coding: utf-8 -*-
from .. import HAS_OPENCV, HAS_PYMONGO
if HAS_OPENCV:
    from .image import Image
    from .metadata import Metadata
    from .analyzer import Analyzer
    from .tape import Tape, TapeAnalyzer
# from .data import DatasetNumpy




# -*- coding: utf-8 -*-

from skimage.color import rgb2gray
import cv2
import numpy as np
import matplotlib.pylab as plt
from skimage import exposure, filters
from pathlib import Path



IMAGE_EXTENSIONS = ['.png', '.bmp', '.dib', '.jpeg', 
                    '.jpg', '.jpe', '.jp2', '.webp',
                    '.pbm', '.pgm', '.ppm', '.pxm', 
                    '.pnm', '.sr', '.ras', '.tiff',
                    '.tif', '.exr', '.hdr', '.pic']


def rotate_image(image, angle):
    """
    Rotates the image by angle degrees

    Parameters
    ----------
        angle : float
            Angle of rotation.

    Returns
    -------
        None.

    """
    
    image_center = tuple(np.array(image.shape[1::-1]) / 2)
    rot_mat = cv2.getRotationMatrix2D(image_center, angle, 1.0)
    image = cv2.warpAffine(
        image, rot_mat, image.shape[1::-1], flags=cv2.INTER_LINEAR)
    return image

def gaussian_blur(image, window=(15, 15)):
    """
    This method applies Gaussian Blur filter to the image. 

    Parameters
    ----------
    window : tuple int, optional
        The window in which the gaussian blur is going to be applied.
        The default is (15,15).

    Returns
    -------
    None.

    """
    image = cv2.GaussianBlur(image, window, 0)
    return image

def split_v(image, pixel_index=None,pick_side='L', flip=True):
    """
    This method splits the image in 2 images based on the fraction that is 
    given in pixel_index

    Parameters
    ----------
    pixel_index : float, optional
        fraction in which the image is going to be split. The value should
        be a number between zero and one. The default is 0.5.
    pick_side : str, optional
        The side in which will over write the image in the class. The 
        default is 'L'.

    Returns
    -------
    None.

    """
    width = image.shape[1]
    
    if pixel_index is None:
        pixel_index = width//2
    if pixel_index<1:
        pixel_index = int(width*pixel_index)
    if pick_side == 'L':
        image = image[:, 0:pixel_index]
    else : 
        if flip:
            image = np.fliplr(image[:,pixel_index:width])
        else:
            image = image[:, pixel_index:width]
    return image

def to_gray(image: np.ndarray, mode='SD') -> np.ndarray:
    """Gray Scale image of the input image.
    
    modes: 'BT.470' and 'BT.709'
    SD 'BT.470' : Y = 0.299 R + 0.587 G + 0.114 B
    HD 'BT.709' : Y = 0.2125 R + 0.7154 G + 0.0721 B

    Returns
    -------
    gray_scale : cv2 object
        Gray Scale image of the input image.

    """
    if len(image.shape) == 2:
        return image
    elif len(image.shape) == 3:
        if image.shape[2] == 1:
            return image
        elif image.shape[2] == 3:
            if mode == 'SD':
                return cv2.cvtColor(image,cv2.COLOR_BGR2GRAY)
            elif mode == 'HD':
                return rgb2gray(image)
            
def to_rbg(image):
    if len(image.shape) == 3 and image.shape[2] == 3:
        return image
    else:
        return cv2.cvtColor(image, cv2.COLOR_GRAY2RGB)

def flip(image):
    return cv2.flip(image, 0)
    
def contours(image, mask_threshold=60):
    """
     A list of pixels that create the contours in the image

    Returns
    -------
    contours : list 
        A list of pixels that create the contours in the image

    """
    masked = get_masked(image, mask_threshold)
    contours,_ = cv2.findContours(
        masked,cv2.RETR_TREE,cv2.CHAIN_APPROX_NONE)
    return contours

def largest_contour(contours):
    """
    A list of pixels forming the contour with the largest area

    Returns
    -------
    contour_max_area : list
        A list of pixels forming the contour with the largest area

    """
    
    max_con = 0
    max_con_arg = 0
    for ic in range(len(contours)):
        if cv2.contourArea(contours[ic]) > max_con :
            max_con = cv2.contourArea(contours[ic])
            max_con_arg = ic
    contour_max_area = contours[max_con_arg]
    return contour_max_area

def remove_background(image: np.array, 
                      contour: np.array, 
                      outside: bool = True,
                      pixel_value: int = 0):
    """Removes the background outside or inside the contour

    Parameters
    ----------
    image : np.array
        _description_
    contour : np.array
        _description_
    outside : bool, optional
        _description_, by default True
    pixel_value : int, optional
        _description_, by default 0
    """
    bkg = np.ones_like(image)*pixel_value
    bkg = cv2.fillPoly(bkg, [contour], (255, 255, 255))
    image = cv2.bitwise_and(image,
                        bkg)
    return image

def get_masked(image, mask_threshold):
        """
        Populates the masked image with the gray scale threshold
        Returns
        
        -------
        None.

        """
        masked = cv2.inRange(image,
                             mask_threshold,
                             255)
        
        return masked
    
def resize(image, size):
    """
    This method resizes the image to the pixel size given.

    Parameters
    ----------
    size : tuple int,
        The target size in which the image is going to be resized. 

    Returns
    -------
    None.

    """
    size = tuple(size)
    assert len(size) == 2, 'Resizing needs two dimensions'
    image = cv2.resize(image,size)
    return image

def exposure_control(image: np.ndarray, 
                     mode:str='equalize_hist', 
                     **kwargs) -> np.ndarray:
    """modifies the exposure

    Parameters
    ----------
    mode : str, optional
        Type of exposure correction. It can be selected from the options:
        ``'equalize_hist'`` or ``'equalize_adapthist'``. 
        `equalize_hist <https://scikit-image.org/docs/stable/api/skimage.exposure.html#equalize-hist>`_ 
        and `equalize_adapthist <https://scikit-image.org/docs/stable/api/skimage.exposure.html#equalize-adapthist>`
        use sk-image. by default 'equalize_hist'
    """
    exps = {'equalize_hist':exposure.equalize_hist,
            'equalize_adapthist':exposure.equalize_adapthist}
    assert mode in exps, 'Mode not valid.'
    image = exps[mode](image, **kwargs)
    return (image*255).astype('uint8')

def apply_filter(image: np.ndarray, mode:str, **kwargs) -> np.ndarray:
    """Applies different types of filters to the image

    Parameters
    ----------
    mode : str
        Type of filter to be applied. The options are
        * ``'meijering'``: <Meijering neuriteness filter https://scikit-image.org/docs/stable/api/skimage.filters.html#skimage.filters.meijering>_,
        * ``'frangi'``: < Frangi vesselness filter https://scikit-image.org/docs/stable/api/skimage.filters.html#skimage.filters.frangi>_,
        * ``'prewitt'``: <Prewitt transform https://scikit-image.org/docs/stable/api/skimage.filters.html#prewitt>_,
        * ``'sobel'``: <Sobel filter https://scikit-image.org/docs/stable/api/skimage.filters.html#skimage.filters.sobel>_,
        * ``'scharr'``: <Scharr transform https://scikit-image.org/docs/stable/api/skimage.filters.html#skimage.filters.scharr>,
        * ``'roberts'``: <Roberts' Cross operator https://scikit-image.org/docs/stable/api/skimage.filters.html#examples-using-skimage-filters-roberts>_,
        * ``'sato'``: <Sato tubeness filter https://scikit-image.org/docs/stable/api/skimage.filters.html#skimage.filters.sato>_.
    """
    flts = {
        'meijering':filters.meijering,
        'frangi': filters.frangi,
        'prewitt': filters.prewitt,
        'sobel': filters.sobel,
        'scharr': filters.scharr,
        'roberts': filters.roberts,
        'sato': filters.sato
    }
    assert mode in flts, 'Filter not valid.'
    if mode == 'roberts':
        if image.ndim != 2:
            print('Cannot apply roberts to color images')
            return
    image = flts[mode](image, **kwargs)
    return (image*255).astype('uint8')

def binerized_mask(image, masked):
    """
    This function return the binarized version of the tape

    Returns
    -------
    2d array of the image
        .

    """
    # image = cv2.bitwise_and(image,
    #                        image,
    #                        mask=masked)
    return cv2.bitwise_and(image,
                           image,
                           mask=masked)
    
def imwrite(fname: str, image: np.array, cmap='gray', **kwargs):
    """save any 2d numpy array (or list) to an image file

    Parameters
    ----------
    fname : str
        flie name to be saved
    image : np.array
        2d numpy array (or list) to be saved
    """
    fname = Path(fname)
    fname.parent.mkdir(exist_ok=True)
    fname = fname.as_posix()
    # cv2.imwrite(fname, image)
    # else:
    plt.imsave(fname, image, cmap=cmap, **kwargs)



# -*- coding: utf-8 -*-
"""
Created on Mon Apr 12 09:36:25 2021

@author: Pedram Tavadze
used PyChemia Database class as a guide
https://github.com/MaterialsDiscovery/PyChemia/blob/master/pychemia/db/db.py
"""
import subprocess
import pymongo
import gridfs
from pathlib import Path
from random import choice
from bson.objectid import ObjectId
from collections.abc import Mapping
from typing import Callable, List, Optional, Union
from ..core import Tape, TapeAnalyzer, Image, Metadata
from ..utils.array_tools import read_bytes_io, write_bytes_io

class ClassMap(Mapping):
    def __init__(self):
        self.mapping = {'material': Tape,
                        'analysis': TapeAnalyzer,
                        'any': Image}
        
    def __contains__(self, x):
        return x in self.mapping

    def __getitem__(self, x):
        if x in self:
            return self.mapping.__getitem__(x)
        else:
            return self.mapping.__getitem__('any')

    def __iter__(self):
        return self.mapping.__iter__()

    def __len__(self):
        return self.mapping.__len__()   
            
class Database:
    def __init__(self,
                 name: str = 'forensicfit',
                 host: str = "localhost",
                 port: int = 27017,
                 username: str = "",
                 password: str = "",
                 verbose: bool = False,
                 **kwargs):

        self.name = name
        self.host = host
        self.port = port

        self.username = username
        self.password = password
        self.verbose = verbose
        
        if len(password) != 0:
            self.password = ":"+password+"@"
        else:
            if len(username) != 0:
                self.username += "@"
        self.uri = "mongodb://%s%s%s:%d" % (username, password, host, port)
        self.client = pymongo.MongoClient(self.uri)
        self.db = self.client[name]
        self.entries = self.db.forensicfit_entries
        self.fs = {}
        self.fs['material'] = gridfs.GridFS(self.db, "material")
        self.fs['analysis'] = gridfs.GridFS(self.db, "analysis")
        self.class_mapping = ClassMap()
        for x in self.collection_names:
            if x not in self.fs:
                self.add_collection(x)
        self.db_info = {"Database Name": self.name,
                        "Host": self.host,
                        "Port": self.port,
                        "User": self.username}
        if self.verbose:
            print("----------------")
            print("connected to:")
            print(self)

    def disconnect(self):
        """Closes the connection with the mongodb Client.

        """
        self.client.close()

    def __str__(self):
        ret = "MongoDB\n"
        ret += "----------------\n"
        for key in self.db_info:
            ret += "{:<15}: {}\n".format(key, self.db_info[key])
        return ret

    def add_collection(self, collection: str):
        self.fs[collection] = gridfs.GridFS(self.db, collection)
        return 

    def exists(self, 
               filter: dict = None, 
               collection: str = None, 
               metadata: Metadata = None) -> ObjectId: #| bool:
        if collection not in self.fs:
            return False
        if metadata is not None:
            collection = collection or metadata['mode']
            filter = {"$and": metadata.to_mongodb_filter()}
        elif filter is None or collection is None:
            raise Exception("Provide metadata or filter and collection")
        ret = self.fs[collection].find_one(filter=filter)
        if ret is not None:
            return ret._id
        else:
            return False

    def insert(self,    
               obj: Union[Image, Tape, TapeAnalyzer],
               ext: str = '.png',
               overwrite: bool = False, 
               skip: bool = False,
               collection : str = None):
        
        collection = collection or obj.metadata['mode']
        filter = {"$and": obj.metadata.to_mongodb_filter()}
        exists = self.exists(filter=filter, 
                             collection=collection,
                             )
        if collection not in self.fs:
            self.add_collection(collection)
        fs = self.fs[collection]
        if overwrite and exists:
            if self.verbose:
                print(f"{obj.metadata.filename} {collection} already exists, overwriting!")
            self.delete(filter, collection)
        elif skip and exists:
            if self.verbose:
                print(f"{obj.metadata.filename} {collection} already exists, skipping!")
            return exists
        metadata = obj.metadata.to_serial_dict
        metadata['ext'] = ext
        filename = obj.metadata.filename
        _id = fs.put(obj.to_buffer(ext),
                     filename = filename,
                     metadata = metadata)
        return _id

    def find(self,
             filter: dict, 
             collection: str = 'analysis',
             ext:str = '.png',
             version: int = -1,
             no_cursor_timeout: bool = False,
             ) -> list:
        Class = self.class_mapping[collection]
        fs = self.fs[collection]

        ret = []
        if self.count_documents(filter, collection) != 0 :
            queries = fs.find(filter=filter, 
                            no_cursor_timeout=no_cursor_timeout).sort("uploadDate", version)
            for iq in queries:
                ret.append(Class.from_buffer(iq.read(), iq.metadata))
        return ret

    def map_to(self,
               func: Callable,
               filter: dict, 
               collection_source: str,
               collection_target,
               verbose: bool=True,
               no_cursor_timeout: bool=False):
        Class = self.class_mapping[collection_source]
        fs = self.fs[collection_source]
        
        queries = fs.find(filter=filter, 
                        no_cursor_timeout=no_cursor_timeout)
        if queries is None:
            print('There are no matching entries to the provided filter')
        for iq in queries:
            obj = Class.from_buffer(iq.read(), iq.metadata)
            ext = obj.metadata['ext']
            if verbose:
                print(iq.filename)
            self.insert(func(obj), ext=ext, collection=collection_target)
            
    def find_one(self, filter = None, collection: str = None) -> object:
        """finds one entry that matches the filter. 
        kwargs must be chosen by filter=

        Parameters
        ----------
        collection: str, optional
            material or analysis, if none is chose it will be selected randomly, by default None

        Returns
        -------
        object
            ForensicFit object e.g. Tape
        """        
        if collection is None :
            collection = choice(list(self.fs))
        Class = self.class_mapping[collection]
        fs = self.fs[collection]
        iq = fs.find_one(filter)
        if iq is not None:
            return Class.from_buffer(iq.read(), iq.metadata)
        else:
            raise ValueError(f'No entry found with the filter: {str(filter)}')
        
    def find_with_id(self, _id: str, collection: str) -> object:
        """Retrieves core object based on the MongoDB _id

        Parameters
        ----------
        _id : str
            MongoDB _id
        collection : str
            'material' or 'analysis'

        Returns
        -------
        object
            ForensicFit object e.g. Tape
        """

        Class = self.class_mapping[collection]
        fs = self.fs[collection]
        iq = fs.find_one({'_id':_id if type(_id) is ObjectId else ObjectId(_id)})
        return Class.from_buffer(iq.read(), iq.metadata)
    
    def filter_with_metadata(self, 
                             inp: List[str], 
                             filter: dict, 
                             collection: str):
        fs = self.fs[collection]
        query = fs.find({'$and':
                        [
                            filter, 
                            {'filename':{'$in':inp}}
                            ]
            })
        db_fnames = [iq.metadata['filename'] for iq in query]
        ret = []
        for i, filename in enumerate(inp):
            if filename in db_fnames:
                ret.append(i)
        return ret
    
    def count_documents(self, filter: dict, collection: str):
        if collection not in self.fs:
            return 0
        fs = self.fs[collection]
        cursor = fs.find()
        return cursor.collection.count_documents(filter=filter)
        
    def export_to_files(self,
                        destination: str,
                        filter: dict,
                        collection: str,
                        ext: str='.png',
                        verbose: bool=True,
                        no_cursor_timeout: bool=False):
        Class = self.class_mapping[collection]
        fs = self.fs[collection]
        queries = fs.find(filter=filter, 
                        no_cursor_timeout=no_cursor_timeout)
        dst = Path(destination)
        dst.mkdir(exist_ok=True)
        for iq in queries:
            obj = Class.from_buffer(iq.read(), iq.metadata)
            ext = obj.metadata['ext']
            path = dst / iq.filename
            if verbose:
                print(path.as_posix())
            obj.to_file(path.with_suffix(ext))

    def drop_collection(self, collection: str):
        for x in ['files', 'chunks']:
            print(f'drop {collection}.{x}')
            self.db.drop_collection(f'{collection}.{x}')
        return

    def delete(self, filter: dict, collection: str):
        fs = self.fs[collection]
        queries = fs.find(filter)
        for iq in queries:
            fs.delete(iq._id)
        return 
        
    def delete_database(self):
        self.client.drop_database(self.name)

    @property
    def collection_names(self):
        return [x.replace(".files", '')for x in self.db.list_collection_names() if 'files' in x]

    @property
    def connected(self):
        try:
            self.client.server_info()  # force connection on a request as the
            return True
        except pymongo.errors.ServerSelectionTimeoutError as err:
            print(err)
            return False
        
    @property
    def server_info(self):
        return self.client.server_info()
    
def dict2mongo_query(inp: dict, previous_key: str = '') -> dict:
    ret = []
    for key in inp:
        if  isinstance(inp[key], dict) :
            if len(inp[key]) != 0 :
                ret.append(dict2mongo_query(inp[key],
                                            previous_key=previous_key+'.'+key,
                                            ))
            else:
                ret.append({previous_key+'.'+key:inp[key]})
        else:
            ret.append({previous_key+'.'+key:inp[key]})
    ret_p = []
    for item_1 in ret:
        if isinstance(item_1, list):
            for item_2 in item_1:
                ret_p.append(item_2)
        else:
            ret_p.append(item_1)
    return ret_p
            
def list_databases(host='localhost',
                   port=27017,
                   username='',
                   password='',
                   ):
    uri = "mongodb://%s%s%s:%d" % (username, password, host, port)
    client = pymongo.MongoClient(uri)
    database_names = client.list_database_names()
    return database_names


def dump(db: Optional[str] = None, 
         host: Optional[str] = None,
         port: Optional[int] = None,
         username: Optional[str] = None,
         password: Optional[str] = None,
         out: Optional[str] = None,
         collection: Optional[str] = None,
         ):
    command = 'mongodump '
    tags = locals()
    for itag in tags:
        if tags[itag] is not None and itag != 'command':
            command += f'--{itag}={tags[itag]} '
    print(command)
    print(subprocess.check_output(command, shell=True))
    
def restore(path: Optional[str] = None,
            db: Optional[str] = None, 
            host: Optional[str] = None,
            port: Optional[int] = None,
            username: Optional[str] = None,
            password: Optional[str] = None,
            collection: Optional[str] = None,
         ):
    command = 'mongorestore '
    tags = locals()
    for itag in tags:
        if tags[itag] is not None and itag not in ['command', 'path']:
            command += f'--{itag}={tags[itag]} '
    command += path
    print(command)
    subprocess.check_output(command, shell=True)



#!/usr/bin/env python3

""" This script takes a metadata file generated by the script 
``create_metadata.py`` and stores all the images listed in that file
stores them in a MongoDB using GridFS.
"""

import os
from typing import Any, List, Dict, Union
from pathlib import Path
import json
import multiprocessing
from multiprocessing import Pool, cpu_count, current_process
import sys
import inspect
import argparse
from unicodedata import name

from numpy import insert
from forensicfit.core import Tape
from forensicfit.database import Database
from forensicfit.utils.image_tools import IMAGE_EXTENSIONS


def get_chunks(files: List[Dict], n_processors: int) -> List[List[Dict]]:
    """divides list of files for multiprocessing

    Parameters
    ----------
    files : List[Dict]
        list of files in a dictionary format
    n_processors : int
        number of processors for Pool

    Returns
    -------
    List[List[Dict]]


    """
    
    ret = [ [] for x in range(n_processors)]
    n_files = len(files)
    for i, ifile in enumerate(files):
        ret[i % n_processors].append(ifile)
    return ret

def worker(args: Dict):
    """inserts images provided in the list to the MongoDB

    connects to a MongoDB, loads the images provided in the args, and 
    inserts them in the DB.

    Parameters
    ----------
    args : Dict
        dictionary containing the files, the database settings, and
        the inser options (skip, overwrite, etc)

    """
    
    files = args['files'] 
    db_settings = args['db_settings']
    insert_options = args['insert_options']
    db = Database(**db_settings)
    for _, entry in enumerate(files):
        file_path = Path(entry['source'])
        if not file_path.exists():
            print(f"{file_path.as_posix()} does not exist")
            return 
        if file_path.suffix not in IMAGE_EXTENSIONS:
            continue
        tape = Tape.from_file(file_path)
        # print(f"{file_path.stem}")
        for key in entry:
            tape.metadata[key] = entry[key]
        db.insert(tape, **insert_options)
        
def store_on_db(
        metadata_file: Union[str, Path],
        db_name: str,
        host:str ='localhost',
        port: int=27017,
        username: str="",
        password:str ="",
        n_processors: int=1,
        overwrite: bool=False,
        skip: bool=True):
    """given a metadata file about images, stores them on a database.

    Given a metadata file generated by create_metadata.py, it divides
    them into chunks, passes them to multiple processors, and inserts
    them on a MongoDB

    Parameters
    ----------
    metadata_file : Union[str, Path]
        Metadata file generated by ``create_metadata.py``
    db_name : str
        Name of the MongoDB to be stored on.
    host : str
        Address to the host of the MongoDB server.
    port : int
        Port of the MongoDB host
    username : str
        User name for the MongoDB host
    password : str
        Password for the MongoDB host
    n_processors : int
        Number of processors to be used.
    overwrite : bool
        If the file already exists, overwrite or create a new one
    skip : bool
        If the file already exists on the DB, skip or not.

    """

    if n_processors > cpu_count():
        n_processors = cpu_count()
        
    with open(metadata_file, 'r') as rf:
        metadata = json.load(rf)
    chunks = get_chunks(metadata, n_processors)
    
    db_settings = dict(name=db_name,
                       host=host,
                       port=port,
                       username=username,
                       password=password)
    insert_options = dict(overwrite=overwrite,
                          skip=skip)
    
    args = [{
            'files': x,
            'db_settings': db_settings,
            'insert_options': insert_options,
        } for x in chunks]
    with Pool(n_processors) as p:
        p.map(worker, iterable=args)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-dbn', '--database-name', 
                           dest='db_name',
                           help='The name of the database.')
    parser.add_argument('-dbh', '--database-host', 
                           dest='db_host',
                           default='localhost',
                           help='The name of the host for the database server.')
    parser.add_argument('-dbp', '--database-port', 
                           dest='db_port',
                           type=int,
                           default=27017,
                           help='The port of the database server.')

    parser.add_argument('-i', '--input',
                                dest='input',
                                help=('Path to the metadata.json file that' 
                                      'contains path all the files. This file ' 
                                      'can be generated by create_metadata.py '
                                      'script')
                                )
    parser.add_argument('-np', '--n-processors', 
                        dest='n_processors',
                        type=int,
                        help='Number of processors available for this task.',
                        default=1)
    parser.add_argument('--overwrite',
                        dest='overwrite',
                        help=('To overwrite the existing entries if they already exist'),
                        action=argparse.BooleanOptionalAction,
                        type=bool,
                        default=False)
    parser.add_argument('--skip',
                        dest='skip',
                        help=('To overwrite the existing entries if they already exist'),
                        action=argparse.BooleanOptionalAction,
                        type=bool,
                        default=False)
    args = parser.parse_args()
    store_on_db(
        metadata_file=args.input,
        db_name=args.db_name,
        host=args.db_host,
        port=args.db_port,
        n_processors=args.n_processors,
        overwrite=args.overwrite,
        skip=args.skip)



# -*- coding: utf-8 -*-
from .. import HAS_OPENCV
from .. import HAS_PYMONGO
if HAS_PYMONGO:
    from . import array_tools
if HAS_OPENCV:
    from . import image_tools
from . import plotter
from .general import copy_doc


# -*- Coding: utf-8 -*-
"""
Created on Sun Jun 28 14:11:02 2020

@author: Pedram Tavadze
"""
import warnings
import numpy.typing as npt
from pathlib import Path
import numpy as np
import matplotlib.pyplot as plt
import math
from ..utils import image_tools
from . import Image, Analyzer
from . import HAS_OPENCV
if not HAS_OPENCV:
    print("To enable analyzer please install opencv")
else:
    import cv2

class Tape(Image):
    def __init__(self,
                 image: npt.ArrayLike,
                 label: str = None,
                 surface: str = None,
                 stretched: bool=False,
                 **kwargs):
        """Tape is a class created for tape images to be preprocessed for 
        Machine Learning. This Class detects the edges, auto crops the image 
        and returns the results in 3 different method coordinate_based, 
        bin_based and max_contrast. 
        """
        assert type(image) is np.ndarray, "image arg must be a numpy array"
        assert image.ndim in [2, 3] , "image array must be 2 dimensional"
        super().__init__(image, **kwargs)
        self.metadata['flip_h'] = False
        self.metadata['flip_v'] = False
        self.metadata['split_v'] = {
            "side": None, "pixel_index": None}
        self.metadata['label'] = label
        self.metadata['material'] = 'tape'
        self.metadata['surface'] = surface
        self.metadata['stretched'] = stretched
        self.metadata['mode'] = 'material'
        
    def split_v(self, side, correct_tilt=True, pixel_index=None):
        self.values['original_image'] = self.image.copy()
        if pixel_index is None:
            tape_analyzer = TapeAnalyzer(self, correct_tilt=correct_tilt, auto_crop=True)
            x = tape_analyzer.boundary[:, 0]
            pixel_index = int((x.max()-x.min())/2+x.min())
            self.image = tape_analyzer['image']
            self.metadata['image_tilt'] = tape_analyzer.metadata.image_tilt
            self.metadata['tilt_corrected'] = tape_analyzer.metadata.tilt_corrected
        self.image = image_tools.split_v(
            self.image, pixel_index, side)
        self.values['image'] = self.image
        self.metadata['split_v'] = {
            "side": side, "pixel_index": pixel_index}
        self.metadata['resolution'] = self.shape
        



class TapeAnalyzer(Analyzer):
    def __init__(self,
                 tape: Tape = None,
                 mask_threshold: int=60,
                 gaussian_blur: tuple=(15, 15),
                 n_divisions: int=6,
                 auto_crop: bool=False,
                 correct_tilt: bool=False,
                 padding: str='tape',
                 remove_background: bool=True):
        """
        

        Parameters
        ----------
        tape : TYPE, optional
            DESCRIPTION. The default is None.
        mask_threshold : TYPE, optional
            DESCRIPTION. The default is 60.
        gaussian_blur : TYPE, optional
            DESCRIPTION. The default is (15, 15).
        n_divisions : TYPE, optional
            DESCRIPTION. The default is 6.
        auto_crop : TYPE, optional
            DESCRIPTION. The default is True.
        calculate_tilt : TYPE, optional
            DESCRIPTION. The default is True.
        verbose : TYPE, optional
            DESCRIPTION. The default is True.

        Returns
        -------
        None.

        """
        super().__init__()
        self.debug = {}
        self.metadata['n_divisions'] = n_divisions
        self.metadata['gaussian_blur'] = gaussian_blur
        self.metadata['mask_threshold'] = int(mask_threshold)
        self.metadata['remove_background'] = remove_background
        self.metadata['padding'] = padding
        self.metadata["analysis"] = {}
        if tape is not None:
            self.image = tape.image
            self.metadata += tape.metadata
            self.metadata['resolution'] = self.image.shape
            self.preprocess()
            self.metadata['cropped'] = auto_crop
            self.metadata['tilt_corrected'] = correct_tilt
            if correct_tilt:
                angle = self.get_image_tilt()
                self.image = image_tools.rotate_image(self.image, angle)
                self.metadata.tilt_corrected = True
            if auto_crop:
                self.get_image_tilt()
                self.auto_crop_y()
                self.metadata.cropped = True
            if correct_tilt or auto_crop:
                self.preprocess()
            if remove_background:
                self.image = self['masked']
            
        return

    def preprocess(self):
        """
        
        Parameters
        ----------
        calculate_tilt : TYPE, optional
            DESCRIPTION. The default is True.
        auto_crop : TYPE, optional
            DESCRIPTION. The default is True.

        Returns
        -------
        None.

        """
        image = self.image
        if np.average(image) > 200:
            image = 255 - self.image
        # if self.metadata.gaussian_blur is not None:
        #     image = image_tools.gaussian_blur(
        #         self.image, window=self.metadata.gaussian_blur)
        gray = image_tools.to_gray(image, mode='SD')
        image = image_tools.gaussian_blur(
                gray, window=self.metadata.gaussian_blur)
        contours = image_tools.contours(image,
                                        self.metadata.mask_threshold)
        largest_contour = image_tools.largest_contour(contours)
        self.metadata['boundary'] = largest_contour.reshape(-1, 2)
            
    def flip_v(self):
        self.image = np.fliplr(self.image)
        self.metadata['boundary'] = np.array(self.metadata['boundary'])
        self.metadata['boundary'][:, 0] = self.metadata['boundary'][:, 0]*-1 + self.image.shape[1]
        if 'coordinate_based' in self.metadata['analysis']:
            coords  = np.array(self.metadata['analysis']['coordinate_based']['coordinates'])
            slopes  = np.array(self.metadata['analysis']['coordinate_based']['slopes'])
            coords[:, 0] *= -1
            coords[:, 0] += self.image.shape[1]
            slopes[:, 1] += slopes[:, 0]*self.image.shape[1]
            slopes[:, 0] *= -1
            # slopes[:, 1] += self.image.shape[1]
            self.metadata['analysis']['coordinate_based']['coordinates'] = coords
            self.metadata['analysis']['coordinate_based']['slopes'] = slopes
        if 'bin_based' in self.metadata['analysis']:
            dynamic_positions = np.array(self.metadata['analysis']['bin_based']['dynamic_positions'])
            for i, pos in enumerate(dynamic_positions):
                for ix in range(2):
                    dynamic_positions[i][0][ix] = pos[0][ix]*-1 + self.image.shape[1]
            dynamic_positions[:, 0, [0, 1]] = dynamic_positions[:, 0, [1, 0]]
            
            self.metadata['analysis']['bin_based']['dynamic_positions'] = dynamic_positions
        self.metadata['flip_v'] =  not(self.metadata['flip_v'])

    def flip_h(self):
        self.image = np.flipud(self.image)
        self.metadata['boundary'] = np.array(self.metadata['boundary'])
        self.metadata['boundary'][:, 1] = self.metadata['boundary'][:, 1]*-1 + self.image.shape[0]
        if 'coordinate_based' in self.metadata['analysis']:
            coords = np.array(self.metadata['analysis']['coordinate_based']['coordinates'])
            slopes  = np.array(self.metadata['analysis']['coordinate_based']['slopes'])
            coords[:, 1] *= -1
            coords[:, 1] += self.image.shape[0]
            slopes *= -1
            slopes[:, 1] += self.image.shape[0]
            self.metadata['analysis']['coordinate_based']['coordinates'] = coords
            self.metadata['analysis']['coordinate_based']['slopes'] = slopes
        if 'bin_based' in self.metadata['analysis']:
            dynamic_positions = np.array(self.metadata['analysis']['bin_based']['dynamic_positions'])
            for i, pos in enumerate(dynamic_positions):
                for ix in range(2):
                    dynamic_positions[i, 1, ix] = pos[1, ix]*-1 + self.image.shape[0]
            dynamic_positions[:, 1, [0, 1]] = dynamic_positions[:, 1, [1, 0]]
            dynamic_positions = np.flipud(dynamic_positions)
            self.metadata['analysis']['bin_based']['dynamic_positions'] = dynamic_positions
        self.metadata['flip_h'] = not(self.metadata['flip_h'])
        
    def load_metadata(self):
        """
        
        Returns
        -------
        None.

        """
        self.metadata["xmin"] = int(self.xmin)
        self.metadata["xmax"] = int(self.xmax)
        self.metadata["x_interval"] = int(self.x_interval)
        self.metadata["ymin"] = int(self.ymin)
        self.metadata["ymax"] = int(self.ymax)
        self.metadata["y_interval"] = int(self.ymax-self.ymin)
        return 

    @classmethod
    def from_dict(cls, image, metadata):
        """
        
        Parameters
        ----------
        cls : TYPE
            DESCRIPTION.
        values : TYPE
            DESCRIPTION.

        Raises
        ------
        Exception
            DESCRIPTION.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        if image is None:
            raise Exception(
                "The provided image was empty. Maybe change the query criteria")
        cls = cls()
        cls.image = image
        cls.metadata.update(metadata)
        return cls

    def get_image_tilt(self, plot: bool=False):
        """
        

        Parameters
        ----------
        plot : TYPE, optional
            DESCRIPTION. The default is False.

        Returns
        -------
        angle_d : TYPE
            DESCRIPTION.

        """
        figsize=(16, 9)
        stds = np.ones((self.metadata.n_divisions-2, 2))*1000
        conditions_top = []
        # conditions_top.append([])
        conditions_bottom = []
        # conditions_bottom.append([])
        
        boundary = self.boundary
        y_min = self.ymin
        y_max = self.ymax

        x_min = self.xmin
        x_max = self.xmax
        m_top = []
        # m_top.append(None)
        m_bottom = []
        # m_bottom.append(None)
        if plot:
            _ = plt.figure(figsize=figsize)
            ax = plt.subplot(111)
            self.plot_boundary(color='black', ax=ax)
            ax.invert_yaxis()
        for idivision in range(self.metadata.n_divisions-2):
            
            y_interval = y_max-y_min
            cond1 = boundary[:, 1] > y_max-y_interval/self.metadata.n_divisions
            cond2 = boundary[:, 1] < y_max+y_interval/self.metadata.n_divisions

            x_interval = x_max-x_min
            cond3 = boundary[:, 0] >= x_min+x_interval/self.metadata.n_divisions*(idivision+1)
            cond4 = boundary[:, 0] <= x_min + \
                x_interval/self.metadata.n_divisions*(idivision+2)

            cond_12 = np.bitwise_and(cond1, cond2)
            cond_34 = np.bitwise_and(cond3, cond4)
            cond_and_top = np.bitwise_and(cond_12, cond_34)

            # This part is to rotate the images

            if sum(cond_and_top) == 0:
                conditions_top.append([])
                m_top.append(None)
                continue
            if plot:
                ax.plot(boundary[cond_and_top][:, 0],
                         boundary[cond_and_top][:, 1], linewidth=3)
            m_top.append(np.polyfit(
                boundary[cond_and_top][:, 0], boundary[cond_and_top][:, 1], 1)[0])
            std_top = np.std(boundary[cond_and_top][:, 1])
            stds[idivision, 0] = std_top
            conditions_top.append(cond_and_top)

        for idivision in range(self.metadata.n_divisions-2):

            cond1 = boundary[:, 1] > y_min-y_interval/self.metadata.n_divisions
            cond2 = boundary[:, 1] < y_min+y_interval/self.metadata.n_divisions

            x_interval = x_max-x_min
            cond3 = boundary[:, 0] >= x_min+x_interval/self.metadata.n_divisions*(idivision+1)
            cond4 = boundary[:, 0] <= x_min + \
                x_interval/self.metadata.n_divisions*(idivision+2)

            cond_12 = np.bitwise_and(cond1, cond2)
            cond_34 = np.bitwise_and(cond3, cond4)
            cond_and_bottom = np.bitwise_and(cond_12, cond_34)
            if sum(cond_and_bottom) == 0:
                conditions_bottom.append([])
                m_bottom.append(None)
                continue
            if plot:
                ax.plot(
                    boundary[cond_and_bottom][:, 0], boundary[cond_and_bottom][:, 1], linewidth=3)
                ax.set_xlim(0, self.image.shape[1])

            m_bottom.append(np.polyfit(
                boundary[cond_and_bottom][:, 0], boundary[cond_and_bottom][:, 1], 1)[0])

            std_bottom = np.std(boundary[cond_and_bottom][:, 1])

            stds[idivision, 1] = std_bottom
            conditions_bottom.append(cond_and_bottom)
        arg_mins = np.argmin(stds, axis=0)
        m_top[arg_mins[0]] = m_bottom[arg_mins[1]] if m_top[arg_mins[0]] is None else m_top[arg_mins[0]]
        m_bottom[arg_mins[1]] = m_top[arg_mins[0]] if m_bottom[arg_mins[1]] is None else m_bottom[arg_mins[1]]
        m = np.average([m_top[arg_mins[0]], m_bottom[arg_mins[1]]])
        cond_and_top = conditions_top[arg_mins[0]]
        cond_and_bottom = conditions_bottom[arg_mins[1]]
        if plot:
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot(111)
            ax.plot(boundary[:, 0], boundary[:, 1], color='black')
            ax.scatter(boundary[cond_and_top][:, 0],
                        boundary[cond_and_top][:, 1], color='blue')
            ax.scatter(boundary[cond_and_bottom][:, 0],
                        boundary[cond_and_bottom][:, 1], color='red')
            ax.set_xlim(0, self.image.shape[1])
            ax.invert_yaxis()
            
        top = boundary[cond_and_top][:, 1]
        bottom = boundary[cond_and_bottom][:, 1]
        self.metadata.crop_y_top = np.min(top) if len(top) != 0 else self.ymax
        self.metadata.crop_y_bottom = np.max(bottom) if len(bottom) !=0 else self.ymin
        if np.min(stds, axis=0)[0] > 10:
            self.metadata.crop_y_top = y_max
        if np.min(stds, axis=0)[1] > 10:
            self.metadata.crop_y_bottom = y_min
            
        angle = np.arctan(m)
        angle_d = np.rad2deg(angle)
        self.metadata.image_tilt = angle_d
        return angle_d

    @property
    def xmin(self):
        """
        X coordinate of minimum pixel of the boundary

        Returns
        -------
        xmin : int
            X coordinate of minimum pixel of the boundary

        """
        xmin = self.boundary[:, 0].min()
        return xmin

    @property
    def xmax(self):
        """
        X coordinate of minimum pixel of the boundary

        Returns
        -------
        xmax : int
            X coordinate of minimum pixel of the boundary

        """
        xmax = self.boundary[:, 0].max()
        return xmax

    @property
    def x_interval(self):
        """
        interval of the coordinates in X direction

        Returns
        -------
        x_interval : int
            interval of the coordinates in X direction

        """
        return self.xmax - self.xmin

    @property
    def ymin(self):
        """
        Y coordinate of minimum pixel of the boundary

        Returns
        -------
        ymin : int
             Y coordinate of minimum pixel of the boundary

        """
        ymin = self.boundary[:, 1].min()
        return ymin

    @property
    def ymax(self):
        """
        Y coordinate of maximum pixel of the boundary

        Returns
        -------
        ymax : int
             Y coordinate of maximum pixel of the boundary

        """
        ymax = self.boundary[:, 1].max()
        return ymax

    def auto_crop_y(self):
        """
        This method automatically crops the image in y direction (top and bottom)


        Parameters
        ----------
        calculate_tilt : TYPE, optional
            DESCRIPTION. The default is False.

        Returns
        -------
        None.

        """
        self.image = self.image[int(
            self.metadata.crop_y_bottom):int(self.metadata.crop_y_top), :]

    def _get_points_from_boundary(self,
                                  x_0: float,
                                  x_1: float,
                                  y_0: float,
                                  y_1: float) -> npt.ArrayLike:
        coordinates = self['boundary']
        cond_1 = coordinates[:, 0] >= x_0
        cond_2 = coordinates[:, 0] <= x_1
        cond_x = np.bitwise_and(cond_1, cond_2)
        cond_1 = coordinates[:, 1] >= y_0
        cond_2 = coordinates[:, 1] <= y_1
        cond_y = np.bitwise_and(cond_1, cond_2)
        cond = np.bitwise_and(cond_x, cond_y)
        return coordinates[cond]

    def get_coordinate_based(self,
                             n_points: int=64,
                             x_trim_param: int=6) -> None:
        """
        This method returns the data of the detected edge as a set of points 
        in 2d plain with (x,y)

        Parameters
        ----------
        n_points : TYPE, optional
            DESCRIPTION. The default is 1024.
        x_trim_param : TYPE, optional
            DESCRIPTION. The default is 2.
        plot : TYPE, optional
            DESCRIPTION. The default is False.

        Returns
        -------
        data : TYPE
            DESCRIPTION.

        """
        was_flipped = self.metadata['flip_v']
        if was_flipped:
            self.flip_v()
        x_min = self.xmin
        x_max = self.xmax
        x_interval = x_max-x_min
        boundary = self.boundary

        # cond_12 = np.bitwise_and(cond1, cond2)
        x_0 = x_min
        x_1 = x_min+x_interval/x_trim_param*0.95
        cond1 = boundary[:, 0] >= x_0
        cond2 = boundary[:, 0] <= x_1
        cond_12 = np.bitwise_and(cond1, cond2)
        # cond_12 = cond1
        data = np.zeros((n_points, 2))
        stds = np.zeros((n_points,))
        slopes = np.zeros((n_points, 2))
        y_min = self.ymin
        y_max = self.ymax
        y_interval = y_max - y_min
        edge = boundary[cond_12]
        self.metadata['edge_x_std'] = float(np.std(edge[:, 0]))
        self.metadata['edge_x_range'] = int(edge[:, 0].max()-edge[:, 0].min())
        self.debug['cb_points'] = []
        for i_point in range(0, n_points):
            y_start = y_min+i_point*(y_interval/n_points)
            y_end = y_min+(i_point+1)*(y_interval/n_points)
            points = self._get_points_from_boundary(x_0, x_1, y_start, y_end)
            
            if len(points) == 0 and x_trim_param != 1:
                if was_flipped:
                    self.flip_v()
                return self.get_coordinate_based(n_points=n_points, 
                                                 x_trim_param=x_trim_param-1)
            self.debug['cb_points'].append(points)
            data[i_point, :2] = np.average(points, axis=0)
            stds[i_point] = np.std(points[:, 0])
            slopes[i_point, :] = np.polyfit(
                x=points[:, 0], 
                y=points[:, 1],
                deg=1,
                full=True)[0]
        self.metadata['analysis']['coordinate_based'] = {
            "n_points": n_points,
            "x_trim_param": x_trim_param,
            'coordinates': data,
            'stds': stds,
            'slopes': slopes}
        if was_flipped:
            self.flip_v()
        return self['coordinate_based']

    def get_bin_based(self,
                      window_background=50,
                      window_tape=1000,
                      dynamic_window=True,
                      size=None,
                      n_bins=10,
                      overlap=0,
                      border: str = 'avg',
                      ):
        """
        This method returns the detected edge as a set of cropped images from 
        the edge. The number if images is defined by n_bins. The goal is 
        to try to match the segmentation to the wefts of the tape. 
        In the future this method will try to detected the wefts automatically.


        Parameters
        ----------
        window_background : int, optional
            Number of pixels to be included in each segment in the background 
            side of the image. The default is 20.
        window_tape : int, optional
            Number of pixels to be included in each segment in the tape side 
            of the image. The default is 300.
        dynamic_window : TYPE, optional
            Whether the windows move in each segment. The default is False.
        size : 2d tuple integers, optional
            The size of each segment in pixels. The default is (300,30).
        n_bins : int, optional
            Number of segments that the tape is going to divided into. The 
            default is 4.
        plot : bool, optional
            Whether or not to plot the divided image. The default is False.


        Returns
        -------
        An array of 2d numpy arrays with images of each segment.

        """
        was_flipped = self.metadata['flip_v']
        if was_flipped:
            self.flip_v()
        overlap = abs(overlap)
        boundary = self.boundary
        x_min = self.xmin
        x_max = self.xmax

        y_min = self.ymin
        y_max = self.ymax

        x_start = x_min-window_background
        x_end = min(x_min+window_tape, self.image.shape[1])

        seg_len = (y_max-y_min)//(n_bins)
        seg_len = (y_max-y_min)/(n_bins)

        segments = []
        dynamic_positions = []
        y_end = y_min
        for iseg in range(0, n_bins):
            y_start = y_end
            y_end = math.ceil(y_min+(iseg+1)*seg_len)
            if dynamic_window:
                cond1 = boundary[:, 1] >= y_start - overlap
                cond2 = boundary[:, 1] <= y_end + overlap
                cond_and = np.bitwise_and(cond1, cond2)
                x_slice = boundary[cond_and, 0]
                if border == 'min' or iseg in [0, n_bins-1]:
                    loc = x_slice.min()
                else:
                    cond_x = x_slice.min() + 500
                    loc = np.average(x_slice[x_slice < cond_x])
                x_start =  loc - window_background
                x_end = loc + window_tape
                if self.image.shape[1] < x_end:
                    diff = self.image.shape[1] - x_end
                    x_start += diff
                    x_end = self.image.shape[1]
            
            ys = y_start - overlap
            ye = y_end + overlap
            
            dynamic_positions.append(
                [[int(x_start), int(x_end)], [int(ys), int(ye)]])

        metadata = {"dynamic_positions": dynamic_positions,
                    "n_bins": n_bins,
                    "overlap": overlap,
                    "dynamic_window": dynamic_window,
                    "window_background": window_background,
                    "window_tape": window_tape,
                    "size": size}

        self.metadata['analysis']['bin_based'] = metadata
        if was_flipped:
            self.flip_v()
        return dynamic_positions

    def get_max_contrast(self,
                         window_background: int=100,
                         window_tape: int=600):
        """
        This method returns the detected image as a black image with only the 
        boundary being white.

        Parameters
        ----------
        window_background : int, optional
            Number of pixels to be included in each segment in the background side of the image. The default is 20.
        window_tape : int, optional
            Number of pixels to be included in each segment in the tape side of the image. The default is 300.
        plot : bool, optional
            Whether or not to plot the divided image. The default is False.

        Returns
        -------
        edge_bw : 2d numpy array
            A black image with all the pixels on the edge white.

        """
        
        self.metadata['analysis']['max_contrast'] = {
            "window_background": window_background,
            "window_tape": window_tape}
        return 

    def __getattr__(self, name):
        return self[name]

    def __getitem__(self, x):
        if x == 'image':
            return self.image
        elif x == 'original_image':
            return self.original_image
        elif x == 'masked':
            return image_tools.remove_background(self.image, 
                                                 self.largest_contour)
        elif x == 'gray_scale':
            return image_tools.to_gray(self.image)
        elif x == 'binarized':
            return image_tools.binerized_mask(
                self['image'], self['masked'])
        elif x == 'largest_contour':
            return self.boundary.reshape(-1, 1, 2)
        elif x == 'boundary':
            return np.asarray(self.metadata['boundary'])
        elif x == 'coordinate_based':
            return self.metadata[
                'analysis']['coordinate_based']
        elif x == 'edge_bw':
            return  cv2.drawContours(
                np.zeros_like(self.image), 
                [self.boundary], 
                0, (255,)*(len(self.image.shape)) , 2)
        elif x == 'max_contrast':
            ret = self['edge_bw']
            x_min = self.xmin
            window_background = self.metadata[
                'analysis']['max_contrast']['window_background']
            window_tape = self.metadata[
                'analysis']['max_contrast']['window_tape']
            x_start = x_min-window_background
            x_end = x_min+window_tape
            pad_x_start = -1*(min(x_start, 0))
            pad_x_end = -1*(min(ret.shape[1]-x_end, 0))
            x_start = max(0, x_start)
            x_end = min(ret.shape[1], x_end)
            ret = ret[:, x_start:x_end]
            ret = np.pad(
                ret,
                ((0, 0),)
                + ((pad_x_start, pad_x_end),) 
                + ((0, 0),)*(len(ret.shape)-2),
                'constant', 
                constant_values=(0,)
                )
            
            return ret
        elif 'bin_based' in x:
            if 'coordinate_based' in x:
                ret = []
                self.debug['cbbb_points'] = []
                n_bins = self.metadata.analysis['bin_based']['n_bins']
                n_points = self.metadata.analysis[
                    'coordinate_based']['n_points']//n_bins
                if self.metadata.analysis['bin_based']['overlap']>50:
                    n_points+=2
                for x, y in self.metadata.analysis[
                    'bin_based']['dynamic_positions']:
                    data = np.zeros((n_points, 2))
                    stds = np.zeros((n_points,))
                    slopes = np.zeros((n_points, 2))
                    y_min = y[0]
                    y_max = y[1]
                    pad_y_min = -1*(min(y_min, 0))
                    pad_y_max = -1*(min(self.image.shape[0]-y_max, 0))
                    y_min = max(y_min, 0)
                    y_max = min(self.image.shape[0], y_max)
                    if self.metadata.padding == 'tape':
                        y_min -= pad_y_max
                        y_max += pad_y_min
                    y_interval = y_max - y_min
                   
                    if n_points > len(self._get_points_from_boundary(
                        x[0], x[1],
                        y_min, y_max
                    )):
                        print('The number of points per bin is smaller than'
                              ' the number of points in this bin.'
                              'Consider decreasing the numbers of points for '
                              'the coordinate based method (n_points).')
                    for i_point in range(n_points):
                        y_start = y_min+i_point*(y_interval/n_points)
                        y_end = y_min+(i_point+1)*(y_interval/n_points)
                        points = []
                        ratios = np.linspace(1, 2, 20)
                        j = 0
                        while len(points) == 0:
                            assert j < len(ratios), ("No points in this window, " 
                                                    "increase the tape window"
                                                    f" {self.metadata.filename}-{self.metadata.split_v['side']}")
                            points = self._get_points_from_boundary(
                                x[0], 
                                x[1]*ratios[j],
                                y_start, 
                                y_end)
                            j += 1
                        
                        
                        self.debug['cbbb_points'].append(points)
                        data[i_point, :2] = np.average(points, axis=0)
                        stds[i_point] = np.std(points[:, 0])
                        slopes[i_point, :] = np.polyfit(
                            x=points[:, 0], 
                            y=points[:, 1], 
                            deg=1, 
                            full=True)[0]
                    ret.append({
                        'coordinates': data,
                        'stds': stds,
                        'slopes': slopes})
                if self.metadata.analysis['bin_based']['overlap']>50:
                    for i, item in enumerate(ret):
                        for key in item:
                            ret[i][key] = np.delete(item[key], [0, -1], 0)
                return ret
            elif x == 'bin_based':
                image = self['image']
            elif 'max_contrast' in x:
                image = self['edge_bw']
            ret = []
            bin_based =  self.metadata['analysis']['bin_based']
            dynamic_positions = np.array(bin_based['dynamic_positions'])
            if bin_based['n_bins'] > 3:
                delta_y = int(np.diff(dynamic_positions[:, 1][1:-2]).mean())
            else:
                delta_y = int(np.diff(dynamic_positions[:, 1]).mean())
            for seg in dynamic_positions:
                x_start, x_end = seg[0]
                y_start, _ = seg[1]
                y_end = y_start + delta_y
                pad_y_start = -1*(min(y_start, 0))
                y_start = max(y_start, 0)
                pad_y_end = -1*(min(self.image.shape[0]-y_end, 0))
                y_end = min(self.image.shape[0], y_end)
                pad_x_start = -1*(min(x_start, 0))
                x_start = max(x_start, 0)
                if self.metadata.padding == 'black':
                    img = image[y_start:y_end, x_start:x_end]
                    # the (len(img.shape)-1) is to adjust to gray scale and rgb
                    img = np.pad(
                        img,
                        ((pad_y_start, pad_y_end),) 
                        + ((pad_x_start, 0),)
                        + ((0, 0),)*(len(img.shape)-2),
                        'constant', 
                        constant_values=(0, )
                        )
                elif self.metadata.padding == 'tape':
                    y_start -= pad_y_end
                    y_end += pad_y_start
                    img = image[y_start:y_end, x_start:x_end]
                ret.append(img)
            n = [x.shape for x in ret]
            if len(set(n)) > 1:
                print(self.metadata.path)
                print(set(n))
            return np.asarray(ret)
        else:
            raise ValueError(f"TapeAnalyzer does not have attrib {x}")    
        




#!/usr/bin/env python3

"""This module will contain a parallelizd script to generate bin based 
data for ML
"""

__author__ = 'Pedram Tavadze'
__date__ = "20220727"
__status__ = 'Production'

import argparse
from multiprocessing import Pool
from pathlib import Path
from typing import List, Dict, Union, Any

import forensicfit as ff

import json
import numpy as np
import pandas as pd
import sys
if ff.HAS_PYMONGO:
    import gridfs
    import pymongo
    from bson.objectid import ObjectId
    from gridfs.grid_file import GridOut


def get_item(name: str, 
             lookup: Any,
             args: argparse.Namespace) -> ff.core.TapeAnalyzer: 
    """Gets a scan from the source database

    Parameters
    ----------
    name : str
        file name
    db : forensicfit.db.Database
        The mongodb object to get the file from

    Returns
    -------
    forensicfit.core.TapeAnalyzer
        analyzed tape scan as a forensicfit object
    """
    # name = name.decode("utf-8")
    fname = f"{name[:-2]}"
    side = name[-1]
    if isinstance(lookup, ff.db.Database):
        criteria = {"$and":[]}
        criteria['$and'].append({'metadata.cropped':args.auto_crop})
        criteria['$and'].append({'metadata.gaussian_blur':list(args.gaussian_blur)})
        criteria['$and'].append({'metadata.mask_threshold':args.mask_threshold})
        criteria['$and'].append({'metadata.n_divisions':args.n_divisions})
        criteria['$and'].append({'metadata.cropped':args.auto_crop})
        criteria['$and'].append(
            {'metadata.analysis.bin_based.dynamic_window':args.dynamic_window}
            )
        criteria['$and'].append({'metadata.analysis.bin_based.n_bins':args.n_bins})
        criteria['$and'].append(
            {'metadata.analysis.bin_based.window_background':args.window_background}
            )
        criteria['$and'].append(
            {'metadata.analysis.bin_based.window_tape':args.window_tape}
            )
        db = lookup
        criteria['$and'].append({'filename':fname})
        _id = db.exists(collection='analysis', filter=criteria)
        if _id:
            tape_analyzer = db.find_with_id(_id, 'analysis')
            if len(tape_analyzer['bin_based']) == 0:
                db.delete(collection='analysis', filter=criteria)
                tape_analyzer = get_item(name, db, args)
            return tape_analyzer
        else:
            criteria = {'filename':fname}
            n_docs = db.count_documents(filter=criteria, collection='material')
            if n_docs == 0:
                return
            else:
                criteria = {"$and":[
                    {'filename':fname}, 
                    {'metadata.modified':True}]
                            }
                if db.count_documents(filter=criteria, collection='material') == 0 :
                    criteria = {'filename':fname}
            tape = db.find_one(filter=criteria, collection='material')
    elif isinstance(lookup, dict):
        key = fname
        if fname+'_mod' in lookup:
            key += '_mod'
        if key in lookup:
            path = Path(lookup[key]["source"])
            if path.exists():
                tape = ff.core.Tape.from_file(path)
            else:
                return
        else:
            return
    tape.split_v(side=side, correct_tilt=args.correct_tilt)
    tape.resize(dpi=tuple(args.dpi))
    tape_analyzer = ff.core.TapeAnalyzer(tape,
                                        mask_threshold=args.mask_threshold,
                                        gaussian_blur=args.gaussian_blur,
                                        n_divisions=args.n_divisions,
                                        auto_crop=args.auto_crop,
                                        padding=args.padding,
                                        )
    tape_analyzer.get_bin_based(window_background=args.window_background,
                                window_tape=args.window_tape,
                                dynamic_window=args.dynamic_window,
                                n_bins=args.n_bins, overlap=args.overlap,
                                border=args.border)
    if args.coordinate_based:
        tape_analyzer.get_coordinate_based(n_points=args.n_points)
    if args.exposure_control is not None:
        tape_analyzer.exposure_control(mode=args.exposure_control)
    if args.apply_filter is not None:
        tape_analyzer.apply_filter(mode=args.apply_filter)
    return tape_analyzer

def preprocess(entry: dict, 
               lookup: Any,
               args: argparse.Namespace):
    rotation_map = {'Backing' : 'Rotation?', 'Scrim' : 'Rotation?.1'}
    rot = ['', '-rotated']
    for tape in ['Tape 1', 'Tape 2']:
        for side in ['Backing', 'Scrim']:
            fname = entry[f'{tape} ({side})']
            tape_analyzer = get_item(fname, lookup, args)
            if tape_analyzer is None:
                continue
            rotated = 0
            if tape == 'Tape 1' and entry[rotation_map[side]]:
                tape_analyzer.flip_h()
                rotated = 1
            if args.bin_based:
                bins = tape_analyzer['bin_based']
                for ibin, b in enumerate(bins):
                    image = ff.core.Image(b)
                    if args.color:
                        image.convert_to_rgb()
                    else:
                        image.convert_to_gray()
                    if args.output_shape is not None:
                        image.resize(args.output_shape)
                    image.metadata.update(tape_analyzer.metadata.to_serial_dict)
                    image.metadata['bin'] = ibin
                    image.metadata['index'] = entry['idx']
                    image.metadata['filename'] = f"{fname}-{ibin}"
                    if args.cln_name is not None:
                        db = lookup
                        db.insert(image,
                                ext=args.ext,
                                collection=args.cln_name)
                    elif args.output is not None:
                        path = Path(args.output)
                        path /= f'{fname}{rot[rotated]}'
                        path /= f"{ibin}{args.ext}"
                        image.to_file(path)
            if args.max_contrast:
                bins = tape_analyzer['bin_based+max_contrast']
                for ibin, b in enumerate(bins):
                    image = ff.core.Image(b)
                    if args.color:
                        image.convert_to_rgb()
                    else:
                        image.convert_to_gray()
                    if args.output_shape is not None:
                        image.resize(args.output_shape)
                    image.metadata.update(tape_analyzer.metadata.to_serial_dict)
                    image.metadata['bin'] = ibin
                    image.metadata['index'] = entry['idx']
                    image.metadata['filename'] = f"{fname}-mxc-{ibin}"
                    if args.cln_name is not None:
                        db = lookup
                        db.insert(image,
                                ext='.png', 
                                collection=args.cln_name)
                    elif args.output is not None:
                        path = Path(args.output)
                        path /= f'{fname}{rot[rotated]}'
                        path /= f"{ibin}-mxc{args.ext}"
                        image.to_file(path)
            if args.coordinate_based:
                bins = tape_analyzer['bin_based+coordinate_based']
                ret = {}
                if args.output is not None:
                    for ibin, b in enumerate(bins):
                        ret[f'bin_{ibin}'] = ff.utils.array_tools.serializer(b)
                    
                    path = Path(args.output)
                    path /= f'{fname}{rot[rotated]}'
                    path.mkdir(exist_ok=True)
                    path /= f"coordinate_based+bin_based.json"
                    with open(path, 'w') as wf:
                        json.dump(ret, wf, indent=2)
                ret = ff.utils.array_tools.serializer(
                    tape_analyzer['coordinate_based'])
                if args.output is not None:
                    path = Path(args.output)
                    path /= f'{fname}{rot[rotated]}'
                    path.mkdir(exist_ok=True)
                    path /= f"coordinate_based.json"
                    with open(path, 'w') as wf:
                        json.dump(ret, wf, indent=2)
    return

def get_chunks(entries: List[Dict], n_processors: int) -> List:
    ret = [ [] for x in range(n_processors)]
    for i, entry in enumerate(entries):
        ret[i % n_processors].append(entry)
    return ret

def worker(args: Dict[str, List]):
    entries = args['entries'] 
    args = args['parsed_args']
    if args.db_name is not None:
        lookup = ff.db.Database(name=args.db_name, 
                                host=args.db_host,
                                port=args.db_port)
    elif args.input is not None:
        with open(args.input, 'r') as rf:
            metadata = json.load(rf)
            lookup = {}
            for x in metadata:
                name = x["filename"]
                if x['modified']:
                    name += '_mod'
                lookup[name] = x
    if args.output is not None:
        path = Path(args.output)
        path.mkdir(exist_ok=True)
    for _, entry in enumerate(entries):
        preprocess(entry, lookup, parsed_args)
    if args.db_name is not None:
        lookup.disconnect()
        

if __name__ == '__main__':
    version = [int(x[:2]) for x in sys.version.replace(']','').replace('[','').split('.')][2]
    parser = argparse.ArgumentParser()
    parser.add_argument('-dbn', '--database-name', 
                           dest='db_name',
                           help='The name of the database.')
    parser.add_argument('-dbh', '--database-host', 
                           dest='db_host',
                           default='localhost',
                           help='The name of the host for the database server.')
    parser.add_argument('-dbp', '--database-port', 
                           dest='db_port',
                           type=int,
                           default=27017,
                           help='The port of the database server.')
    # TODO add username and password arguments for database.
    parser.add_argument('-cln', '--collection-name', 
                           dest='cln_name',
                           help=('The name of the target collection in which the'
                                 'data will be stored'))
    parser.add_argument('-i', '--input',
                                dest='input',
                                help=('Path to the metadata.json file that' 
                                      'contains path all the files. This file ' 
                                      'can be generated by create_metadata.py '
                                      'script')
                                )
    parser.add_argument('-o', '--output',
                        dest='output',
                        help=('Path to the directory for the output '
                              'image')
                        )

    parser.add_argument('-np', '--n-processors', 
                        dest='n_processors',
                        type=int,
                        help='Number of processors available for this task.',
                        default=2)

    parser.add_argument('-nb', '--n-bins',
                        dest='n_bins',
                        type=int,
                        required=True,
                        help='Number of bins to divide the tape in.',
                        )
    parser.add_argument('-nd', '--n-divisions',
                        dest='n_divisions',
                        type=int,
                        default=6,
                        help=('Parameter to find the edge. Number of divisions '
                              'in the x direction'))
    parser.add_argument('-wbg', '--window-background',
                        dest='window_background',
                        type=int,
                        required=True,
                        help=('Number of pixels to include from the background '
                              '(black) of the image. This parameter is measured '
                              'from the edge of the tape towards the background.'
                              ))
    parser.add_argument('-wtp', '--window-tape',
                        dest='window_tape',
                        type=int,
                        required=True,
                        help=('Number of pixels to include from the tape '
                              'of the image. This parameter is measured '
                              'from the edge of the tape towards the tape.'))
    parser.add_argument('--overlap',
                        dest='overlap',
                        type=int,
                        default=100,
                        help=('Number of pixels to include from the adjacent '
                              'tape.'))
    parser.add_argument('-x', '-extension',
                        dest='ext',
                        default='.png',
                        help='The extension of file to be save.')
    parser.add_argument('-s', '--start',
                        dest='start',
                        type=int,
                        default=0,
                        help=('Index to start from in the excel file of the'
                              ' ground truth.'))
    parser.add_argument('-e', '--end',
                        dest='end',
                        type=int,
                        default=-1,
                        help=('Index to end to in the excel file of the'
                              ' ground truth.'))
    parser.add_argument('--excel-files',
                        dest='path_excel',
                        required=True,
                        nargs='+')

    parser.add_argument('--output-shape', 
                        dest='output_shape',
                        type=int,
                        nargs=2,
                        help='The output to be in gray scale or color.'
                        )
    parser.add_argument('--gaussian-blur', 
                        dest='gaussian_blur',
                        type=int,
                        nargs=2,
                        help=('Gaussian blur applied to the image for edge' 
                              ' detection.'),
                        default=[15, 15],
                        )
    parser.add_argument('--mask-threshold', 
                        dest='mask_threshold',
                        default=60,
                        type=int,
                        help=('Threshold for binarization of the image for edge'
                              ' detection.')
                        )
    parser.add_argument('--individual-entry', 
                        dest='individual_entry',
                        help=('files the individual entry from the excel files ' 
                              'and only processes those.'),
                        )
    parser.add_argument('--n-points',
                        dest='n_points',
                        type=int,
                        help='Number of points (in total) for coordinate based',
                        )

    parser.add_argument('--exposure-control',
                        dest='exposure_control',
                        default=None,
                        help=(
                            'Fixes the exposure of the image. options are: '
                            'equalize_hist and equalize_adapthist.'))
    parser.add_argument('--apply-filter',
                        dest='apply_filter',
                        default=None,
                        help=('Applies filter to the images. options are: '
                              'meijering, frangi, prewitt, sobel, scharr, '
                              'roberts, sato.'))
    parser.add_argument('--padding',
                        dest='padding',
                        default='tape',
                        help=('Padding in the y direction for the overlap.'
                              'black or tape'
                              ))
    parser.add_argument('--bin-based-border',
                        dest='border',
                        default='avg',
                        help=('The border used when applying window tape and '
                              'window background. avg or min.'
                              ))
    parser.add_argument('--dpi',
                        dest='dpi',
                        default=None,
                        type=int,
                        nargs=2,
                        help=('Output image, DPI.'))
    if version < 9:
        parser.add_argument('--max-contrast',
                            dest='max_contrast',
                            help=('To generate the maximum contrast for each bin'),
                            action=argparse.BooleanOptionalAction,
                            type=bool,
                            default=False)
        parser.add_argument('--coordinate-based',
                            dest='coordinate_based',
                            help=('To generate the coordinate based for each bin'),
                            action=argparse.BooleanOptionalAction,
                            type=bool,
                            default=False)
        parser.add_argument('--bin-based',
                            dest='bin_based',
                            help=('To generate the bin-based bins'),
                            action=argparse.BooleanOptionalAction,
                            type=bool,
                            default=True)
        parser.add_argument('--color', 
                            dest='color',
                            type=bool,
                            action=argparse.BooleanOptionalAction,
                            help='The output to be in gray scale or color.',
                            default=False,
                            )
        parser.add_argument('--auto-crop',
                            dest='auto_crop',
                            type=bool,
                            default=True,
                            action=argparse.BooleanOptionalAction,
                            help=('To crop the image according to the boundaries in '
                                'the y direction'))
        parser.add_argument('--correct-tilt',
                            dest='correct_tilt',
                            type=bool,
                            default=True,
                            action=argparse.BooleanOptionalAction,
                            help=('To calculate the angle of the image with the '
                                'horizontal line'))
        parser.add_argument('--dynamic-window', 
                            dest='dynamic_window',
                            type=bool,
                            default=True,
                            help='Use a dynamic window when scanning the edge',
                            action=argparse.BooleanOptionalAction,
                            )
    else:
        parser.add_argument('--max-contrast',
                            dest='max_contrast',
                            help=('To generate the maximum contrast for each bin'),
                            action='store_true')
        parser.add_argument('--no-max-contrast',
                            dest='max_contrast',
                            help=('To generate the maximum contrast for each bin'),
                            action='store_false')
        parser.add_argument('--coordinate-based',
                            dest='coordinate_based',
                            help=('To generate the coordinate based for each bin'),
                            action='store_true')
        parser.add_argument('--no-coordinate-based',
                            dest='coordinate_based',
                            help=('To generate the coordinate based for each bin'),
                            action='store_false')                            
        parser.add_argument('--bin-based',
                            dest='bin_based',
                            help=('To generate the bin-based bins'),
                            action='store_true')
        parser.add_argument('--no-bin-based',
                            dest='bin_based',
                            help=('To generate the bin-based bins'),
                            action='store_false')
        parser.add_argument('--color', 
                            dest='color',
                            help='The output to be in gray scale or color.',
                            action='store_true')
        parser.add_argument('--no-color', 
                            dest='color',
                            help='The output to be in gray scale or color.',
                            action='store_false')
        parser.add_argument('--auto-crop',
                            dest='auto_crop',
                            action='store_true',
                            help=('To crop the image according to the boundaries in '
                                'the y direction'))
        parser.add_argument('--no-auto-crop',
                            dest='auto_crop',
                            action='store_false',
                            help=('To crop the image according to the boundaries in '
                                'the y direction'))
        parser.add_argument('--correct-tilt',
                            dest='correct_tilt',
                            action='store_true',
                            help=('To calculate the angle of the image with the '
                                'horizontal line'))
        parser.add_argument('--no-correct-tilt',
                            dest='correct_tilt',
                            action='store_false',
                            help=('To calculate the angle of the image with the '
                                'horizontal line'))
        parser.add_argument('--dynamic-window', 
                            dest='dynamic_window',
                            help='Use a dynamic window when scanning the edge',
                            action='store_true',
                            )
        parser.add_argument('--no-dynamic-window', 
                            dest='dynamic_window',
                            help='Use a dynamic window when scanning the edge',
                            action='store_false',
                            )
        parser.set_defaults(max_contrast=False, 
                            coordinate_based=False,
                            bin_based=True,
                            color=False,
                            auto_crop=True,
                            correct_tilt=True,
                            dynamic_window=True)
    parsed_args = parser.parse_args()
    print('\n'.join(f'{k}={v}' for k, v in vars(parsed_args).items()))
    dfs = [pd.read_excel(x, engine='openpyxl') for x in parsed_args.path_excel]
    df = pd.concat(dfs)
    df = df[[x for x in df.columns if 'Tape' in x or 'Rotation' in x]]
    df['idx'] = np.arange(1, len(df) + 1 )
    df['Rotation?'] = df['Rotation?'].astype(bool)
    df['Rotation?.1'] = df['Rotation?.1'].astype(bool)
    if parsed_args.individual_entry is not None:
        dfs = []
        for col in df.columns:
            dfs.append(df[df[col] == parsed_args.individual_entry])
        df = pd.concat(dfs)
    else:
        df = df[parsed_args.start:parsed_args.end]
    del dfs
    chunks = get_chunks(df.to_dict('records'), parsed_args.n_processors)
    args = [
        {'entries': x,
        'parsed_args': parsed_args} 
        for x in chunks
        ]
    if parsed_args.n_processors == 1:
        worker(args[0])
    else:
        with Pool(parsed_args.n_processors) as p:
            p.map(worker, args)
    output = Path(parsed_args.output or '.')
    output /= f'log_{parsed_args.start}_{parsed_args.end}.json'
    with open(output, 'w') as wf:
        json.dump(parsed_args.__dict__, wf, indent=2)



# -*- coding: utf-8
# import tensorflow as tf

from .. import HAS_TENSORFLOW
import numpy as np
from collections.abc import Mapping
from abc import ABCMeta, abstractmethod
import os

"""
TODO :
  imbalearn
"""

class DatasetNumpy:

    def __init__(self, X :np.array, y: np.array, extra={}, name=''):
        self.X = X
        self.y = y
        self.extra = {key:np.array(extra[key]) for key in extra}
        self.values = {"X": self.X, "y": self.y, "extra": self.extra}
        self.values.update(self.extra)
        self.metadata = {"mode": "data", 'name': name}
        self.name = name

        self._train_indicies = None
        self._test_indicies = None
        self.shuffle()

    @property
    def shape(self):
        ret = {key:self.extra[key].shape for key in self.extra}
        ret['X'] = self.X.shape
        ret['y'] = self.y.shape
        return ret

        
    def shuffle(self, train_size=0.8):
        """


        Parameters
        ----------
        train_size : TYPE, optional
            DESCRIPTION. The default is 0.8.

        Returns
        -------
        None.

        """
        indicies = np.random.randint(0, self.ndata, (self.ndata))
        train_length = int(round(0.8*self.ndata, 0))
        self._train_indicies = indicies[:train_length]
        self._test_indicies = indicies[train_length:]

        
    def balance(self):
        print("balancing the data")
        indicies = self.indicies
        nclasses = {key:len(indicies[key]) for key in indicies}
        smallest_class = min(nclasses, key=nclasses.get)
        print("class with the smallest data points {} with {} points".format(smallest_class, nclasses[smallest_class]))
        for cl in indicies:
            if cl == smallest_class:
                continue
            else:
                nremove = nclasses[cl] - nclasses[smallest_class]
                print(nremove, cl)
                idxs = np.random.choice(indicies[cl], size=nremove, replace=False)
                # indicies[cl][np.random.randint(nclasses[cl], size=nremove)]

                self.X = np.delete(self.X, idxs, 0)
                self.y = np.delete(self.y, idxs, 0)
                for key in self.extra:
                    self.extra[key] = np.delete(self.extra[key], idxs, 0)
        self.shuffle()
        
    @property
    def train_X(self):
        """


        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        return self.X[self._train_indicies]

    @property
    def train_y(self):
        """


        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        return self.y[self._train_indicies]

    @property
    def test_X(self):
        """


        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        return self.X[self._test_indicies]

    @property
    def test_y(self):
        """


        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        return self.y[self._test_indicies]

    def save(self, filename=''):
        """


        Parameters
        ----------
        filename : TYPE, optional
            DESCRIPTION. The default is ''.

        Returns
        -------
        None.

        """
        if ".npy" in filename:
            filename = filename.split(".")[0]
        if len(filename) == 0:
            filename = self.name
        np.save("{}_X.npy".format(filename), self.X)
        np.save("{}_y.npy".format(filename), self.y)
        print("file {} saved".format(filename))

        if len(filename) == 0:
            filename = self.name
        if ".npy" not in filename:
            filename +='.npz'
        i = 1
        fp = filename
        while os.path.isdir(fp):
            fp = filename+str(i)
            i += 1
        filename = fp
        np.savez(filename, **self.values)
        print("files saved at {}".format(filename))
        self.shape

    @classmethod
    def load(cls, filename):
        if ".npz" not in filename:
            filename +='.npz'
        values = np.load(filename, allow_pickle=True)
        X = values['X']
        y = values['y']
        extra = {}
        for ifile in values.files:
            if ifile in ["X", "y"]:
                continue
            else:
                extra[ifile] = values[ifile]
        cls = DatasetNumpy(X, y, extra=extra, name=filename[:-4])
        print('loaded {}'.format(filename))
        cls.shape
        return cls
        
    
    @property
    def ndata(self):
        """
        

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        if self.X is not None:
            return len(self.X)
        else:
            return 0

    @classmethod
    def load(cls, filename=''):
        """
        

        Parameters
        ----------
        cls : TYPE
            DESCRIPTION.
        filename : TYPE, optional
            DESCRIPTION. The default is ''.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        if ".npy" in filename:
            filename = filename.split(".")[0]
        X = np.load("{}_X.npy".format(filename))
        y = np.load("{}_y.npy".format(filename))
        cls = DatasetNumpy(X, y, name=filename)
        return cls

    @property
    def inputs(self):
        """
        

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        return self.X

    @property
    def outputs(self):
        """
        

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        return self.y

    @property
    def nclass(self):
        return len(self.classes)

    @property
    def classes(self):
        return np.unique(self.y)

    @property
    def indicies(self):
        ret = {}
        for cl in self.classes:
            ret[cl] = np.where(self.y == cl)[0]
        return ret
    
    def __str__(self):
        shape = self.shape
        to_print = 'Dataset Numpy, {}\n'.format(self.name)
        to_print += '============================\n'
        to_print += 'number of classes {}\n'.format(self.nclass)
        indicies = self.indicies
        for cl in self.classes:
            to_print += 'number of elements in class {} = {}\n'.format(cl, len(indicies[cl]))
        to_print += '============================\n'        
        for key in shape:
            to_print += "shape of {:>18} = {}\n".format(key, str(shape[key]))
        
        return to_print

    def __contains__(self, x):
        return x in self.values

    def __getitem__(self, x):
        return self.values.__getitem__(x)

    def __iter__(self):
        return self.values.__iter__()

    def __len__(self):
        return self.values.__len__()

    def __add__(self, new):
        X = np.append(self.X, new.X, axis=0)
        y = np.append(self.y, new.y, axis=0)
        extra = {}
        for key in self.extra:
            extra[key] = np.append(self.extra[key], new.extra[key], axis=0)    
        name = self.name
        return DatasetNumpy(X, y, name)
        return DatasetNumpy(X, y, extra, name)





# -*- coding: utf-8 -*-
import importlib.util
import sys
from .version import __version__
def has_package(name):
    spec = importlib.util.find_spec(name)
    if name in sys.modules:
        ret = True
    elif spec is not None:
        ret = True
    else:
        ret = False
    return ret



HAS_OPENCV = has_package("cv2")
HAS_SKIMAGE = has_package("skimage")
HAS_PYMONGO = has_package("pymongo")
HAS_TENSORFLOW = has_package("tensorflow")


from . import core
from . import utils
if HAS_OPENCV and HAS_PYMONGO:
    from . import database as db
    # from .script_process_dir import process_directory
    # from .store_on_db import store_on_db
    # from .from_excel import from_excel




# -*- coding: utf-8 -*-

import io
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import List, Union

import cv2
from forensicfit.utils import plotter
import numpy as np
from matplotlib import pylab as plt
import matplotlib as mpl
from matplotlib.axes import Axes
from scipy.stats import norm

from forensicfit import utils

from ..utils import copy_doc, image_tools, plotter
from .metadata import Metadata

IMAGE_EXTENSIONS = image_tools.IMAGE_EXTENSIONS

class Analyzer:
    
    __metaclass__ = ABCMeta
    """Class containing all future analyzers
    """
    
    def __init__(self, **kwargs):
        """
        
        Returns
        -------
        None.

        """
        
        self.image = None
        self.values = {}
        self.metadata = Metadata({'mode': 'analysis', 
                                  'label': None, 
                                  'material': None})
        self.metadata.update(kwargs)

    # @copy_doc(image_tools.exposure_control)
    def exposure_control(self, mode:str='equalize_hist', **kwargs):
        self.original_image = self.image
        self.image = image_tools.exposure_control(self.image, mode, **kwargs)
        if self.metadata['remove_background']:
            self.image = image_tools.remove_background(self.image, 
                                                       self.largest_contour)
        self.metadata['exposure_control'] = mode
        if len(kwargs) != 0:
            for key in kwargs:
                self.metadata[key] = kwargs[key]
        return

    # @copy_doc(image_tools.apply_filter)
    def apply_filter(self, mode:str, **kwargs):
        self.original_image = self.image
        self.image = image_tools.apply_filter(self['gray_scale'], mode, **kwargs)
        self.metadata['filter'] = mode
        if len(kwargs) != 0:
            for key in kwargs:
                self.metadata[key] = kwargs[key]
        return

    def resize(self, size: tuple = None, dpi: tuple = None):
        if dpi is None and size is not None:
            self.image = image_tools.resize(self.image, size)
            self.values['image'] = self.image
            self.metadata['resize'] = size
            self.metadata['resolution'] = self.image.shape
        elif dpi is not None and 'dpi' in self.metadata:
            dpi = np.array(dpi, dtype=np.int_)
            dpi_old = np.array(self.metadata.dpi, dtype=np.int_)
            ratio = dpi/dpi_old
            size = np.flip((np.array(self.shape)[:2]*ratio).round().astype(int))
            self.image = image_tools.resize(self.image, size)
            self.values['image'] = self.image
            self.metadata['resize'] = size
            self.metadata['resolution'] = self.image.shape
            self.metadata['resolution'] = self.image.shape
            self.metadata['dpi'] = dpi

    def plot_boundary(self, 
                      savefig: Union[str, Path] = None, 
                      color: str='red', 
                      ax: Axes = None, 
                      show: bool=False):
        """
        This function plots the detected boundary of the image. 

        Parameters
        ----------
        savefig : str, optional
            path to save the plot. The default is None.
        
        color : str, optional
            Color of the boundary. The default is 'r'

        ax : str, optional
            Re. The default is None

        show: bool
            Controls whether to show the image. Default is False

        

        Returns
        -------
        None.

        """
        if ax is None:
            plt.figure(figsize = (16, 9))
            ax = plt.subplot(111)
        ax.plot(self.boundary[:, 0], self.boundary[:, 1], c = color)
        if savefig is not None:
            plt.savefig(savefig)
        elif show:
            plt.show()
        return ax

    def plot(self,
             which: str, 
             cmap: str='gray',
             zoom: int=4,
             savefig: str = None, 
             ax: Axes or List[Axes]= None, 
             show: bool = False,
             mode: str = None, 
             **kwargs):
        """
        Parameters
        ----------
        which : TYPE
            DESCRIPTION.
        cmap : TYPE, optional
            DESCRIPTION. The default is 'gray'.
        savefig : TYPE, optional
            DESCRIPTION. The default is None.
        ax : TYPE, optional
            DESCRIPTION. The default is None.
        reverse_x : TYPE, optional
            DESCRIPTION. The default is False.
        savefig : str
            Path and name to save the image. Default is None

        ax : str
            Path and name to save the image. Default is None

        revere_x: bool

        show: bool
            Controls whether to show the image. Default is False

        plot_gaussian: bool
            plots the gaussian smearing method. Default is False

        plot_errorbar: bool
            plots the gaussian smearing method. Default is False
        Returns
        -------
        None.

        """

        if which == "coordinate_based":
            if ax is None:
                figsize = plotter.get_figure_size(self.metadata['dpi'],
                                                  self.shape[:2], 
                                                  zoom)
                fig = plt.figure(figsize=figsize)
                ax = fig.add_subplot(111)
            coordinates = self['coordinate_based']['coordinates']
            stds = self['coordinate_based']['stds']
            slopes = self['coordinate_based']['slopes']
            ax = utils.plotter.plot_coordinate_based(coordinates,
                                                slopes,
                                                stds,
                                                mode,
                                                ax, **kwargs)
            ax.set_xlim(0, self['image'].shape[1])
            # ax.set_xlim(0, self.image.shape[1])
            # ax.set_ylim(0, self.image.shape[0])
            ax.invert_yaxis()
        elif which == 'boundary':
            ax = self.plot('image', cmap=cmap, ax=ax)
            ax = self.plot_boundary(ax=ax)
            ax_ = ax
        elif which in ['bin_based+coordinate_based',
                       'coordinate_based+bin_based']:
            if mode == 'individual_bins':
                dynamic_positions = np.array(self.metadata[
                    'analysis']['bin_based']['dynamic_positions'])
                xmin = min(dynamic_positions[:, 0, 0])
                xmax = max(dynamic_positions[:, 0, 1])
                
                
                
                n_bins = self.metadata['analysis']['bin_based']['n_bins']
                if ax is None:
                    figure = plt.figure(figsize=(5, 2*n_bins))
                    ax = figure.subplots(
                        n_bins, 1,
                        sharex=True, 
                        gridspec_kw={'hspace':2e-2})
                
                elif isinstance(ax, list):
                    assert len(ax) >= n_bins, 'Number of Axes provided ' \
                        "smaller than the number of bins"
                if n_bins == 1: ax=[ax]

                for i, i_bin in enumerate(self[which]):
                    coordinates = i_bin['coordinates']
                    stds = i_bin['stds']
                    slopes = i_bin['slopes']
                    ax[i] = utils.plotter.plot_coordinate_based(coordinates,
                                                    slopes,
                                                    stds,
                                                    mode,
                                                    ax[i], **kwargs)
                    dy = coordinates[1, 1] - coordinates[0, 1]
                    y_min, y_max = min(coordinates[:, 1]), max(coordinates[:, 1])
                    # ax[i].xaxis.set_visible(False)
                    # ax[i].yaxis.set_visible(False)
                    # ax[i].set_ylim(y_min-dy, y_max+dy)
                    y1 = dynamic_positions[i][1][0]
                    y2 = dynamic_positions[i][1][1]
                    x1 = dynamic_positions[i][0][0]
                    x2 = dynamic_positions[i][0][1]
                    ax[i].set_xlim(0, self.xmax)
                    ax[i].set_ylim(y2, y1)
                    if not ax[i].yaxis_inverted():
                        ax[i].invert_yaxis()
                ax_ = ax[-1]
                
                
                # ax_.set_xlim(xmin, xmax)
                # ax.set_ylim(xmin, xmax)

            else:
                if ax is None:
                    plt.figure(figsize=(16, 9))
                    ax = plt.subplot(111)
                for i, i_bin in enumerate(self[which]):
                    coordinates = i_bin['coordinates']
                    stds = i_bin['stds']
                    slopes = i_bin['slopes']
                    ax = utils.plotter.plot_coordinate_based(coordinates,
                                                    slopes,
                                                    stds,
                                                    mode,
                                                    ax, **kwargs)


                ax.set_ylim(0, self.image.shape[0])
                ax.set_xlim(0, self.image.shape[1])
                ax.invert_yaxis()
                
            
        elif which in [
            'bin_based',
            'bin_based+max_contrast',
            'max_contrast+bin_based']:
            if mode == 'individual_bins':
                dynamic_positions = self.metadata[
                    'analysis']['bin_based']['dynamic_positions']
                n_bins = self.metadata['analysis']['bin_based']['n_bins']
                if ax is None:
                    figure = plt.figure(figsize=(10, 20))
                    ax = figure.subplots(
                        n_bins, 1,
                        gridspec_kw={'hspace':2e-2})
                
                elif isinstance(ax, list):
                    assert len(ax) >= n_bins, 'Number of Axes provided ' \
                        "smaller than the number of bins"
                if n_bins == 1: ax=[ax]
                
                bins = self[which]
                for i, seg in enumerate(bins):
                    y1 = dynamic_positions[i][1][0]
                    y2 = dynamic_positions[i][1][1]
                    x1 = dynamic_positions[i][0][0]
                    x2 = dynamic_positions[i][0][1]
                    # ax[i].set_facecolor('black')
                    ax[i].imshow(seg, cmap=cmap, extent=(x1, x2, y1, y2))
                    # ax[i].set_xlim(x1, x2)
                    # ax[i].set_ylim(y1, y2)
                    ax[i].xaxis.set_visible(False)
                    ax[i].yaxis.set_visible(False)
                    # ax[i].imshow(seg, cmap=cmap)
                ax_ = ax[-1]
            else:
                if ax is None:
                    plt.figure(figsize=(16, 9))
                    ax = plt.subplot(111)
                dynamic_positions = self.metadata[
                    'analysis']['bin_based']['dynamic_positions']
                colors = [
                    'red', 'blue', 'green', 'cyan', 'magenta'
                    ]*len(dynamic_positions)
                styles = [
                    'solid', 'dashed', 'dotted', 'dashdot'
                    ]*len(dynamic_positions)
                xs = []
                for i, seg in enumerate(dynamic_positions):
                    y1 = seg[1][0]
                    y2 = seg[1][1]
                    x1 = seg[0][0]
                    x2 = seg[0][1]
                    xs.append(x1)
                    xs.append(x2)
                    ax.plot([x1, x1], [y1, y2], color=colors[i], 
                            linestyle=styles[i], linewidth=1)
                    ax.plot([x2, x2], [y1, y2], color=colors[i], 
                            linestyle=styles[i], linewidth=1)
                    ax.plot([x1, x2], [y1, y1], color=colors[i], 
                            linestyle=styles[i], linewidth=1)
                    ax.plot([x1, x2], [y2, y2], color=colors[i], 
                            linestyle=styles[i], linewidth=1)
                if 'max_contrast' in which:
                    ax = self.plot('edge_bw', ax=ax, cmap=cmap)
                    ax_ = ax
                else:
                    ax = self.plot('image', ax=ax, cmap=cmap)
                    ax_ = ax
        else:
            if ax is None:
                figsize = plotter.get_figure_size(self.metadata['dpi'],
                                                  self.shape[:2], 
                                                  zoom)
                plt.figure(figsize = figsize)
                ax = plt.subplot(111)
            ax.imshow(self[which], cmap=cmap)
            ax_ = ax
            ax.set_xlim(0, self[which].shape[1])
        if 'coordinate_based' not in which:
            ax_.xaxis.set_visible(False)
            ax_.yaxis.set_visible(False)
        if savefig is not None:
            plt.savefig(savefig)
            return ax
        if show:
            plt.show()
        else:
            return ax

    @abstractmethod
    def load_dict(self):
        pass

    @property
    def shape(self):
        return self.image.shape

    @abstractmethod
    def from_dict(self):
        pass

    @classmethod
    def from_buffer(cls, 
                    buffer: bytes, 
                    metadata: dict,
                    ext: str='.png'):
        """receives an io byte buffer with the corresponding metadata and 
        creates the image class

        Parameters
        ----------
        buffer : io.BytesIO
            _description_
        metadata : dict
            _description_
        allow_pickle : bool, optional
            _description_, by default False
        """        
        if 'ext' in metadata:
            ext = metadata['ext']
        if ext in IMAGE_EXTENSIONS:
            image = cv2.imdecode(np.frombuffer(buffer, np.uint8), -1)
            return cls.from_dict(image, metadata)
    
    def to_buffer(self, ext: str = '.png'):
        if ext in IMAGE_EXTENSIONS:
            is_success, buffer = cv2.imencode(ext, self.image)
            output = io.BytesIO(buffer)
        else:
            raise ValueError("Extension not supported")
        return output.getvalue()
        
    def __contains__(self, x):
        return x in self.values

    def __getitem__(self, x):
        return self.values.__getitem__(x)

    def __iter__(self):
        return self.values.__iter__()

    def __len__(self):
        return self.values.__len__()

    def __repr__(self) -> str:
        self.plot(which='boundary', show=True)
        ret = self.metadata.__str__()
        return ret


#!/usr/bin/env python3

""" Searches through all directories, finds files matching the
extension, compares it excel file provided, and returns a json
style lookup table.
"""
__author__ = "Pedram Tavadze"

import pandas as pd
from pathlib import Path
import re
import json
import argparse
from PIL import Image, ExifTags
from typing import Union, Dict, List

Image.MAX_IMAGE_PIXELS = None


def get_files(path: Path, 
              ret: list = {}, 
              ext: str = '.tif',
              avoid: list = []) -> List[Dict[str, Path]]:
    for x in path.iterdir():
        if x.is_file() and x.suffix == ext :
            ret[x.stem] = x
        elif x.is_dir() and x.stem not in avoid:
            get_files(x, ret, ext, avoid)
    return ret

def get_metadata(file_path: Path, key: str) -> Dict[str, str]:
    surface = re.findall('\(([a-zA-Z]*)\)', key)[0]
    if 'stretch' in file_path.as_posix().lower():
        stretched = True
    else:
        stretched = False            
    
    if 'hand torn' in file_path.as_posix().lower():
        separation_method = "Hand Torn"
    elif 'scissor cut' in file_path.as_posix().lower():
        separation_method = "Scissor Cut"
    
    if "medium quality" in file_path.as_posix().lower():
        quality = "Medium Quality"
    elif "high quality" in file_path.as_posix().lower():
        quality = "High Quality"
    elif "low quality" in file_path.as_posix().lower():
        quality = "Low Quality"

    if 'mod' in file_path.as_posix():
        modified = True
    else:
        modified = False
    filename = file_path.stem.replace('_mod','')
    pillow_image = Image.open(file_path)
    dpi = [int(x) for x in pillow_image.info['dpi']]
    pillow_image.close()
    ret = {'filename': filename,
            'surface': surface,
            'quality': quality,
            'separation_method': separation_method,
            'stretched': stretched,
            'modified': modified, 
            'dpi': dpi,
            'source': file_path.absolute().as_posix(),
            }
    
    xr, yr = dpi
    print(f'{filename:<15} | {surface:<10} | {quality:<20} | dpi: {xr}, {yr}')
    return ret

def get_lookup(files_dict: Dict[str, str], 
               df: pd.DataFrame) -> List[Dict]:
    """generates a lookup table for metadata

    Parameters
    ----------
    files_dict : Dict[str, str]
        Dictionary containing metadata about all files found
    df : pd.DataFrame
        Match or non-match excel file

    Returns
    -------
    List[Dict]
        list of all the files that have been mentioned in the pandas
        dataframe provided and their metadata


    """
    lookup = []
    not_exists = []
    for entry in df.to_dict('records'):
        for key in entry:
            if 'Tape' in key:
                name = entry[key][:-2]
                if name in files_dict:
                    file_path = files_dict[name]
                if name+'_mod' in files_dict:
                    file_path = files_dict[name+'_mod']
                lookup.append(get_metadata(file_path, key))    
                if name not in files_dict:
                    not_exists.append(entry[key])
    return lookup, not_exists
    

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=(
        "Searches through all directories, finds files matching the"
        "extension, compares it excel file provided, and returns a json"
        "style lookup table.")
                                     )
    parser.add_argument('-d', '--dpath', 
                        dest='path',
                        required=True,
                        help='Directory to tree search')
    parser.add_argument('-a', '--avoid', 
                        dest='avoid', 
                        nargs='+', 
                        help='Directories to avoid searching',
                        default=[])
    parser.add_argument('-e','--excel-files',
                        dest='path_excel',
                        required=True,
                        nargs='+')
    parser.add_argument('-x','--extension',
                        dest='ext',
                        default='.tif',
                        help='Extension of the files being searched for')
    parser.add_argument('-o', '--output',
                        dest='output',
                        type=str,
                        default='metadata.json',
                        help='Output (metadata) filename',
                        )
    
    args = parser.parse_args()
    
    file_dict = get_files(Path(args.path), ext=args.ext, avoid=args.avoid)
    dfs = [pd.read_excel(x, engine='openpyxl') for x in args.path_excel]
    df = pd.concat(dfs)
    cols = [x for x in df.columns if 'Tape' in x or 'Rotation' in x]
    df = df[cols]
    del dfs
    lookup, not_exists = get_lookup(file_dict, df)
    print("-------------------------------------------------")
    print('The following files do not exist on this storage.')
    for x in not_exists:
        print(x)
    with open(args.output,'w') as wf:
        json.dump(lookup, wf, indent=2)



# -*- coding: utf-8 -*-

import io

import cv2
import numpy as np
from numpy import typing as npt

from .. import HAS_PYMONGO

def serializer(indict: dict) -> dict:
    """Serilizes any given dictionary for mongodb.

    Parameters
    ----------
    indict : dict
        input dictionary

    """
    
    ret = {}
    for key in indict:
        if type(indict[key]) is dict:
            ret[key] = serializer(indict[key])
        elif type(indict[key]) is np.ndarray:
            ret[key] = indict[key].tolist()
        else :
            ret[key] = indict[key]
    return ret
    
def vote_calculator(prediction: npt.ArrayLike) -> npt.ArrayLike:
    n_voters = prediction.shape[1]
    ret = np.zeros(shape=(prediction.shape[0]))
    for i, pred in enumerate(prediction):
        votes = pred.round()
        score = votes.sum()/n_voters
        ret[i] = score
    return np.array(ret)


if HAS_PYMONGO:
    from gridfs.grid_file import GridOut


    def read_bytes_io(obj: GridOut, method: str = 'numpy') -> np.array:
        """reads a binary file stored in mongodb and returns a numpy array

        Parameters
        ----------
        obj : GridOut
            output from a mongodb girdfs file

        Returns
        -------
        np.array
            numpy array containing the information loaded from gridfs file

        """
        if method == 'numpy':
            return np.load(io.BytesIO(obj.read()), allow_pickle=True)
        elif method == 'opencv':
            return cv2.imdecode(np.frombuffer(obj.read().getbuffer(), np.uint8), -1)


    def write_bytes_io(obj: dict, method: str = 'numpy') -> io.BytesIO:
        if method == 'numpy':
            output = io.BytesIO()
            np.savez(output, **obj)
            return output.getvalue()
        elif method == 'opencv':
            is_success, buffer = cv2.imencode(".png", obj)
            output = io.BytesIO(buffer)
            return output.getvalue()




# -*- coding: utf-8 -*-
"""
used PyChemia code output class as a guide  
pychemia/code/codes.py
"""

__all__ = []
__version__ = '1.0'
__author__ = 'Pedram Tavadze'

import io
import pathlib
from abc import ABCMeta, abstractmethod
from collections.abc import Mapping
from pathlib import Path


import cv2
import numpy as np
import numpy.typing as npt
import PIL
from matplotlib import pylab as plt
from matplotlib.axes import Axes
from scipy import ndimage
from skimage import exposure, filters
from typing import Union

from ..utils import copy_doc, image_tools, plotter
from .metadata import Metadata

IMAGE_EXTENSIONS = image_tools.IMAGE_EXTENSIONS
PIL.Image.MAX_IMAGE_PIXELS = None
# mpl.rcParams['image.origin'] = 'lower'

class Image(Mapping):
    __metaclass__ = ABCMeta

    def __init__(self, image: npt.ArrayLike, **kwargs):
        """_summary_

        Parameters
        ----------
        image : np.array
            _description_
        label : str, optional
            _description_, by default None
        """        

        self.image = image
        self.values = {'image': self.image}
        self.metadata = Metadata({'mode': 'image', 
                                  'resolution': self.image.shape})
        self.metadata.update(kwargs)

    @classmethod
    def from_file(cls, filepath: Union[str, pathlib.Path]):

        path = Path(filepath)
        if path.exists():
            image = cv2.imread(path.as_posix())
            pillow_image = PIL.Image.open(path)
            image_info = pillow_image.info
            pillow_image.close()
            if 'icc_profile' in image_info:
                del image_info['icc_profile']
            return cls(image, path=path, filename=path.name, **image_info)
        else:
            raise Exception(f"File {path.as_posix()} does not exist")
        
 
    @classmethod
    def from_buffer(cls, 
                    buffer: bytes, 
                    metadata: dict,
                    ext: str='.npz', 
                    allow_pickle: bool=False):
        """receives an io byte buffer with the corresponding metadata and creates 
        the image class

        Parameters
        ----------
        buffer : io.BytesIO
            _description_
        metadata : dict
            _description_
        allow_pickle : bool, optional
            _description_, by default False
        """
        if 'ext' in metadata:
            ext = metadata['ext']
        if ext == '.npz':
            values = dict(np.load(
                io.BytesIO(buffer), 
                allow_pickle=allow_pickle))
            del metadata['ext']
            return cls.from_dict(values, metadata)
            # cls = Image(values['image'], label=metadata['filename'])
            # cls.metadata = metadata
        elif ext in IMAGE_EXTENSIONS:
            image = cv2.imdecode(np.frombuffer(buffer, np.uint8), -1)
            obj = cls(image)
            obj.metadata.update(metadata)
            return obj

    @classmethod
    def from_dict(cls, values: dict, metadata: dict):
        return cls(values['image'], metadata)
        
                    
    def to_buffer(self, ext: str = '.png'):
        if ext == '.npz':
            output = io.BytesIO()
            np.savez(output, self.values)
        elif ext in IMAGE_EXTENSIONS:
            is_success, buffer = cv2.imencode(ext, self.image)
            output = io.BytesIO(buffer)
        return output.getvalue()

    def to_file(self, filepath: str):
        image_tools.imwrite(filepath, self.image)
        return 
    
    @copy_doc(image_tools.exposure_control)
    def exposure_control(self, mode:str='equalize_adapthist', **kwargs):
        self.image = image_tools.exposure_control(self.image)
        self.metadata['exposure_control'] = mode
        if len(kwargs) != 0:
            for key in kwargs:
                self.metadata[key] = kwargs[key]
        return

    @copy_doc(image_tools.apply_filter)
    def apply_filter(self, mode:str, **kwargs):
        self.image = image_tools.apply_filter(self.image, **kwargs)
        self.metadata['filter'] = mode
        if len(kwargs) != 0:
            for key in kwargs:
                self.metadata[key] = kwargs[key]
        return

    def isolate(self,
                x_start: int = None, 
                x_end: int = None, 
                y_start: int = None, 
                y_end: int = None) -> np.ndarray:
        """isolates a rectangle from the image

        Parameters
        ----------
        x_start : int
            _description_
        x_end : int
            _description_
        y_start : int
            _description_
        y_end : int
            _description_

        Returns
        -------
        np.ndarray
            _description_
        """
        x_start = x_start or 0
        x_end = x_end or self.image.shape[1]
        y_start = y_start or 0
        y_end = y_end or self.image.shape[0]
        return self.image[y_start:y_end, x_start:x_end]
    
    def crop(self,
             x_start: int, 
             x_end: int, 
             y_start: int, 
             y_end: int) -> np.ndarray:
        self.image = self.isolate(**locals())
        self.metadata['resolution'] = self.image.shape
        return
    
    @property
    def shape(self):
        return self.image.shape
        
    def convert_to_gray(self):
        self.image = image_tools.to_gray(self.image)
        self.metadata['resolution'] = self.image.shape
    
    def convert_to_rgb(self):
        self.image = image_tools.to_rbg(self.image)
        self.metadata['resolution'] = self.image.shape


    def rotate(self, angle: float):
        self.image = image_tools.rotate_image(self.image, angle)
        
    def plot(self,
             savefig: str = None, 
             cmap:str='gray',
             ax: Axes = None,
             show: bool=False, 
             zoom: int=4,
             **kwargs):
        """
        

        Parameters
        ----------
        savefig : TYPE, optional
            DESCRIPTION. The default is None.
        cmap : TYPE, optional
            DESCRIPTION. The default is 'gray'.
        ax : TYPE, optional
            DESCRIPTION. The default is None.
        rotate : TYPE, optional
            DESCRIPTION. The default is 0.0.

        Returns
        -------
        None.

        """

        if ax is None:
            
            figsize = plotter.get_figure_size(dpi=self.metadata.dpi, 
                                              image_shape=self.shape[:2],
                                              zoom=zoom)
            plt.figure(figsize=figsize)
            ax = plt.subplot(111)
        image = self.image
        ax.imshow(image, cmap=cmap)
        ax.xaxis.set_visible(False)
        ax.yaxis.set_visible(False)
        plt.tight_layout()
        if savefig is not None:
            plt.savefig(savefig)
        if show:
            plt.show()

        return ax
                
    def show(self, wait=0, savefig = None):
        """
        

        Parameters
        ----------
        wait : TYPE, optional
            DESCRIPTION. The default is 0.
        savefig : TYPE, optional
            DESCRIPTION. The default is None.

        Returns
        -------
        None.

        """
        cv2.imshow(self.label, self.image)
        cv2.waitKey(wait)
        cv2.destroyAllWindows()
        if savefig is not None:
            cv2.imwrite(savefig, self.image)

    def resize(self, size: tuple = None, dpi: tuple = None):
        if dpi is None and size is not None:
            self.image = image_tools.resize(self.image, size)
            self.values['image'] = self.image
            self.metadata['resize'] = size
            self.metadata['resolution'] = self.image.shape
        elif dpi is not None and 'dpi' in self.metadata:
            dpi = np.array(dpi, dtype=np.int_)
            dpi_old = np.array(self.metadata.dpi, dtype=np.int_)
            ratio = dpi/dpi_old
            size = np.flip((np.array(self.shape)[:2]*ratio).round().astype(int))
            self.image = image_tools.resize(self.image, size)
            self.values['image'] = self.image
            self.metadata['resize'] = size
            self.metadata['resolution'] = self.image.shape
            self.metadata['resolution'] = self.image.shape
            self.metadata['dpi'] = dpi

    def flip_h(self):
        self.image = image_tools.flip(self.image)
        self.values['image'] = self.image
        self.metadata['flip_h'] = True
        self.metadata['resolution'] = self.image.shape

    def flip_v(self):
        self.image = np.fliplr(self.image)
        self.values['image'] = self.image
        self.metadata['flip_v'] = True
        self.metadata['resolution'] = self.image.shape

    def copy(self):
        return self.from_dict(self.values, self.metadata)

    def __contains__(self, x):
        return x in self.values

    def __getitem__(self, x):
        return self.values.__getitem__(x)

    def __iter__(self):
        return self.values.__iter__()

    def __len__(self) -> int:
        return self.values.__len__()

    def __repr__(self) -> str:
        self.plot(show=True)
        ret = self.metadata.__str__()
        return ret



from email.mime import image
import itertools
from typing import Any, List, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from .. import HAS_TENSORFLOW
from scipy import stats
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from pathlib import Path
from matplotlib.ticker import MaxNLocator, MultipleLocator

# plt.rcParams["font.family"] = "Arial"
# plt.rc("font", size=11)  # controls default text sizes
# plt.rc("axes", titlesize=14)  # fontsize of the axes title
# plt.rc("axes", labelsize=14)  # fontsize of the x and y labels
# plt.rc("xtick", labelsize=12)  # fontsize of the tick labels
# plt.rc("ytick", labelsize=12)  # fontsize of the tick labels
# plt.rc("legend", fontsize=12)  # legend fontsize


def get_figure_size(dpi: Tuple[int, int],
                    image_shape: Tuple[int, int], 
                    zoom=4,
                    margin=4) -> Tuple[float, float]:
    dpi = np.array(dpi, dtype=np.float_)
    image_shape = np.array(image_shape, dtype=np.float_)
    fig_size = np.flip(image_shape/dpi)*zoom
    # print(fig_size, dpi, image_shape)
    return fig_size+margin
    
def plot_coordinate_based(coordinates: npt.ArrayLike,
                          slopes: npt.ArrayLike,
                          stds: npt.ArrayLike,
                          mode: str = None,
                          ax: Axes = None,
                          plot_slope: bool=True,
                          plot_error_bars: bool=True,
                          plot_edge: bool=True,
                          show: bool=True,
                          dark_bg: bool=True,
                          **kwargs) -> Axes:
    if ax is None:
        fig = plt.figure()
        ax = fig.subplot(111)
    # if mode == "gaussians":
    #     dy = (self.ymax-self.ymin)/n_points
    #     # norm = Normalize(vmin, vmax)
    #     cmap=plt.get_cmap('gray')   
    #     coordinates[:, 1] = np.flip(coordinates[:, 1])
    #     for i, ig in enumerate(coordinates):
    #         x = np.linspace(ig[0]-3*stds[i], ig[0]+3*stds[i])
    #         dx = (x[2]-x[1])
    #         y = np.ones_like(x)*ig[1]
    #         y_prime = norm.pdf(x, ig[0], stds[i])
    #         y_prime /= sum(y_prime)/dx
    #         colors = cmap(y_prime)
    #         y_prime*=dy
    #         ax.fill_between(x, y, y+y_prime, cmap='gray')
    #         ax.scatter(coordinates[:, 0],
    #             coordinates[:, 1],
    #             c='black',
    #             s=0.01)
    if dark_bg:
        ax.set_facecolor('black')
    color=['cyan', 'magenta', 'black', 'orange', 'red', 'blue', 'green']
    if plot_slope:
        dy = coordinates[1, 1] - coordinates[0, 1]
        dy *= 0.7
        for i, iseg in enumerate(slopes):
            m = iseg[0]
            b0 = iseg[1]
            y0 = coordinates[i, 1]
            y_min = y0 - dy/2
            y_max = y0 + dy/2
            y = np.linspace(y_min, y_max, 100)
            x = y/m - b0/m
            ax.plot(x, y, color='C9')
        color=['C1']
    fig = ax.get_figure()
    large_dim = fig.get_size_inches().max()
    marker_size = large_dim/4
    if plot_error_bars:
        ax.errorbar(coordinates[:, 0],
                    coordinates[:, 1],
                    xerr=stds,
                    ecolor='C4',
                    color=np.random.choice(color, size=(1,),)[0],
                    )
    else:
        ax.scatter(coordinates[:, 0],
                   coordinates[:, 1],
                   color=np.random.choice(color, size=(1,),)[0])

    if plot_edge:
        ax.plot(coordinates[:, 0],
                coordinates[:, 1],
                color='C0')
    # ax.set_ylim(min(coordinates[:, 1]),max(coordinates[:, 1]))            
    # xmin = min(coordinates[:, 0])
    # xmax = max(coordinates[:, 0])
    # ax.set_xlim(xmin-abs(xmin)*0.9, xmax+abs(xmax)*1.1)
    return ax
    

def plot_pair(obj_1: Any,
              obj_2: Any,
              text: List[str] = None,
              which: str='boundary',
              mode: str = None,
              savefig: str = None,
              cmap: str='gray',
              show: bool = True,
              figsize: tuple = None,
              labels: List[str] = None,
              title: str = None,
              zoom:int=4,
              **kwargs,
              ) -> Tuple[Figure, Axes]:
    if text is not None:
        n_columns = 3
    else:
        n_columns = 2
    if 'bin_based' in which and mode == 'individual_bins':
        n_bins = max(
            obj_1.metadata['analysis']['bin_based']['n_bins'],
            obj_1.metadata['analysis']['bin_based']['n_bins']
        )
        x = obj_1.shape[0]
        y = obj_1.metadata['analysis']['bin_based']['window_background']
        y += obj_1.metadata['analysis']['bin_based']['window_tape']
        y *= 2
        x *= 2
        figsize = figsize or get_figure_size(obj_1.metadata['dpi'], 
                                             (x, y), 
                                             zoom) #(n_bins/4, 2*n_bins)
        figure = plt.figure(figsize=figsize)
        axes = figure.subplots(
            n_bins, n_columns,
            gridspec_kw={'hspace': 0.0, 'wspace': 0.00})
        ax = [axes[:, 0], axes[:, 1]]
    else:
        figsize = figsize or get_figure_size(obj_1, zoom)
        figure = plt.figure(figsize=figsize)
        axes = figure.subplots(1, n_columns)
        ax = [axes[0], axes[1]]
    for i, obj in enumerate([obj_1, obj_2]):
        axi = obj.plot(which, ax=ax[i], mode=mode, cmap=cmap, **kwargs)
        for a in axi:
            a.set_xticklabels([])
            a.set_yticklabels([])
        if labels is not None:
            axi.set_title(labels[i])
    if title is not None:
        figure.suptitle(title)
    if text is not None:
        for i, tex in enumerate(text):
            axes[i, 2].text(0.5, 0.5, tex)
            axes[i, 2].xaxis.set_visible(False)
            axes[i, 2].yaxis.set_visible(False)
    if show:
        plt.show()
    if savefig is not None:
        plt.savefig(savefig)
    return figure, ax

def plot_pairs(objs: List[Any],
              text: List[str] = None,
              which: str='boundary',
              mode: str = None,
              savefig: str = None,
              cmap: str='gray',
              show: bool = False,
              figsize: tuple = None,
              labels: List[str] = None,
              title: str = None,
              zoom:int=4,
              **kwargs,
              ) -> Tuple[Figure, Axes]:
    n_objs = len(objs)
    if text is not None:
        n_columns = n_objs + 1
    else:
        n_columns = n_objs
    if 'bin_based' in which and mode == 'individual_bins':
        n_bins = max(
            [x.metadata['analysis']['bin_based']['n_bins'] for x in objs]
            )
        overlap = objs[0].metadata['analysis']['bin_based']['overlap']
        X = objs[0].shape[0] + overlap*(n_bins-2)
        Y = 0
        for obj in objs:
            for window in ['window_background', 'window_tape']:
                Y += obj.metadata.analysis['bin_based'][window]
        if text is not None:
            Y += Y/n_objs
        figsize = figsize or get_figure_size(objs[0].metadata.dpi,
                                             (X, Y), 
                                             zoom
                                             )
        figure, axes = plt.subplots(n_bins, n_columns, 
                                    figsize=figsize,
                                    )
        # figure = plt.figure(figsize=figsize)
        #  = figure.subplots(
            # 
        #     gridspec_kw={'hspace': 0.0, 'wspace': 0.0})
        if len(objs) != 2:
            ax = [axes[:, i] for i in range(n_objs)]
        else:
            ax = axes
    else:
        X = objs[0].shape[0]
        Y = objs[0].shape[1]*n_objs
        figsize = figsize or get_figure_size(objs[0].metadata.dpi,
                                             (X, Y), 
                                             zoom)
        figure = plt.figure(figsize=figsize)
        axes = figure.subplots(1, n_columns, 
                               gridspec_kw={'hspace': 0.0, 'wspace': 0.0})
        ax = [axes[i] for i in range(n_objs)]
    print(f'Created a figure with the dimensions: {figsize[0]}, {figsize[1]}')
    for i, obj in enumerate(objs):
        axi = obj.plot(which, ax=ax[i], mode=mode, cmap=cmap, **kwargs)
        # if not isinstance(axi, list): 
        #     axi = [axi]
        for a in axi:
            a.set_xticklabels([])
            a.set_yticklabels([])
        if labels is not None:
            axi.set_title(labels[i])
    if title is not None:
        figure.suptitle(title)
    if text is not None:
        for i, tex in enumerate(text):
            axes[i, -1].text(0.1, 0.1, tex)
            axes[i, -1].xaxis.set_visible(False)
            axes[i, -1].yaxis.set_visible(False)
    plt.tight_layout()
    if savefig is not None:
        plt.savefig(savefig)
    if show:
        plt.show()

    return figure, ax

def plot_confusion_matrix(matrix: np.asarray,
                          class_names: np.asarray,
                          title: str = 'Confusion matrix',
                          cmap: str = 'Blues',
                          normalize: bool = False,
                          savefig: str = None,
                          ax: Axes = None,
                          show: bool = True,
                          colorbar: bool = True):

    accuracy = np.trace(matrix) / float(np.sum(matrix))
    misclass = 1 - accuracy

    cmap = plt.get_cmap(cmap)

    if ax is None:
        plt.figure(figsize=(8, 6))
        ax = plt.subplot(111)
    im = ax.imshow(matrix, cmap=cmap)
    ax.set_title(title)
    if colorbar:
        plt.colorbar(im)

    if class_names is not None:
        tick_marks = np.arange(len(class_names))
        ax.set_xticks(tick_marks, class_names)  # , rotation=45)
        ax.set_yticks(tick_marks, class_names)

    if normalize:
        matrix = matrix.astype('float') / matrix.sum(axis=1)[:, np.newaxis]

    thresh = matrix.max() / 1.5 if normalize else matrix.max() / 2
    for i, j in itertools.product(range(matrix.shape[0]), range(matrix.shape[1])):
        if normalize:
            ax.text(j, i, "{:0.4f}".format(matrix[i, j]),
                     horizontalalignment="center",
                     color="white" if matrix[i, j] > thresh else "black")
        else:
            ax.text(j, i, "{:,}".format(matrix[i, j]),
                     horizontalalignment="center",
                     color="white" if matrix[i, j] > thresh else "black")

    # ax.set_ylabel('True label')
    # ax.set_xlabel('Predicted label\naccuracy={:0.2f}; misclass={:0.2f}'.format(
        # accuracy, misclass))
    if savefig:
        plt.savefig(savefig)
    if show:
        plt.show()
    return ax

def plot_kde_distribution(distribution: np.ndarray,
                            color: str = 'blue',
                            opacity: float = 1.0,
                            label: str = '',
                            ax: Axes = None,
                            savefig: str = None,
                            fill_curve: bool = True,
                            show: bool = True):
    span = np.linspace(distribution.min(), distribution.max())
    kernel = stats.gaussian_kde(distribution)
    estimated_distribution = kernel(span)
    if ax is None:
        plt.figure(16, 9)
        ax = plt.subplot(111)
    ax.plot(span, estimated_distribution, color=color, label=label)
    if fill_curve:
        ax.fill_between(span, estimated_distribution, color=color, alpha=opacity)
    ax.set_xlim(0, 1)
    if savefig:
        plt.savefig(savefig)
    if show:
        plt.show()
    return ax
        
def plot_hist_distribution(distribution: np.ndarray,
                           color: str = 'blue',
                           opacity: float = 1.0,
                           label: str = '',
                           ax: Axes = None,
                           savefig: str = None,
                           show: bool = True):
    
    if ax is None:
        plt.figure(16, 9)
        ax = plt.subplot(111)
    ax.hist(distribution, color=color, label=label, alpha=opacity)
    ax.set_xlim(0, 1)
    if savefig:
        plt.savefig(savefig)
    if show:
        plt.show()
    return ax

if HAS_TENSORFLOW:
    import tensorflow as tf
    def plot_metrics(history: tf.keras.callbacks.History,
                    metrics: List[Union[tf.keras.metrics.Metric, str]],
                    has_valid: bool=True,
                    savefig: str = None):
        """Plots the given 'metric' from 'history'.
        """
        n_plots = len(metrics)
        n_rows = n_plots//2 + n_plots%2
        fig = plt.figure(figsize=((14, n_rows*2+10)))
        axes = fig.subplots(n_rows, 2, sharex=True)
        for i, metric in enumerate(metrics):
            if 'keras' in str(type(metric)):
                name = metric.name
                ylabel = metric.__str__().split('(')[0]
            elif isinstance(metric, str):
                name = metric
                ylabel = metric.capitalize()
            EPOCHS = len(history[name])
            x = np.arange(1, EPOCHS+1)
            ax = axes[i//2, i%2]
            ax.plot(x, history[name], color='blue')
            if has_valid:
                ax.plot(x, history["val_" + name], color='red')
                ax.legend(["train", "validation"])#, loc="upper left")
            ax.xaxis.set_major_locator(MaxNLocator(integer=True))
            ax.xaxis.set_minor_locator(MultipleLocator(1))
            ax.tick_params(axis="x", direction="inout")
            ax.tick_params(which="minor", axis="x", direction="in") 
            ax.tick_params(axis="y", direction="inout")
            ax.set_ylabel(ylabel)
            if i//2 == n_rows - 1:
                ax.set_xlabel("epoch")
            ax.set_xlim(1, EPOCHS)
        plt.tight_layout()
        if savefig is not None:
            savefig = Path(savefig)
            plt.savefig(savefig.absolute().as_posix())
        plt.show()


from typing import Callable

def copy_doc(copy_func: Callable) -> Callable:
    """Use Example: copy_doc(self.copy_func)(self.func) or used as deco"""
    def wrapper(func: Callable) -> Callable:
        func.__doc__ = copy_func.__doc__
        return func
    return wrapper



from setuptools import find_packages, setup
from forensicfit import __version__ as version

with open("requirements.txt") as f:
    requirements = []
    extras = {}
    reqs = f.read().splitlines()
    optional = False
    for req in reqs:
        if len(req) == 0:
            continue
        if req.startswith("#"):
            req = req.replace("#", "").strip()
            if "optional" in req.lower():
                section = req.lower().replace('optional', '').strip()
                extras[section] = []
                optional = True
            else:
                extras[section].append(req)
        else:
            requirements.append(req)

setup(
    name="forensicfit",
    version=version,
    description="A Python library for forensic image comparision.",
    author='Pedram Tavadze',
    author_email='petavazohi@mix.wvu.edu',
    url="https://github.com/romerogroup/ForensicFit",
    license="LICENSE.txt",
    install_requires=requirements,
    extras_require=extras,
    scripts=["scripts/create_metadata.py", 
             'scripts/preprocess_bin_based.py', 
             'scripts/store_on_db.py'],
    packages=find_packages(exclude=["docs", 'sphinx']),
)



__all__ = []
__version__ = '1.0'
__author__ = 'Pedram Tavadze'

# built-ins
from abc import ABCMeta, abstractmethod
from collections.abc import MutableMapping
from pathlib import Path
from ..utils.array_tools import serializer


class Metadata(MutableMapping):
    # __metaclass__ = ABCMeta
    
    def __init__(self, inp: dict = None, **kwargs):
        
        self.mapping = {}
        self.update(inp)

    def __contains__(self, x):
        return x in self.mapping

    def __setitem__(self, key, value) -> None:
        if 'pathlib' in str(type(value)):
            value = value.as_posix()
        self.mapping[key] = value
        if not isinstance(getattr(type(self), key, None), property):
                setattr(self, key, value)
        return 

    def __delitem__(self, key):
        del self.mapping[key]

    def __getitem__(self, x):
        return self.mapping.__getitem__(x)

    def __iter__(self):
        return self.mapping.__iter__()

    def __len__(self):
        return self.mapping.__len__()

    def __repr__(self):
        ret = ''
        for key in self.mapping:
            if key not in ['boundary', 'analysis']:
                value = self.mapping[key] 
                desc = key.capitalize() if key != 'dpi' else key.upper()
                desc = desc.replace('_', ' ')
                desc = desc.replace(' h', ' horizontal')
                desc = desc.replace(' v', ' vertical')
                ret += f'{desc}: {value}\n'
        return ret
    
   
    def __add__(self, new):
        for key in new.mapping:
            if key not in self:
                self[key] = new[key]
        return self
    
    def to_mongodb_filter(self, inp: dict = None, previous_key: str = 'metadata') -> dict:
        inp = inp or self.to_serial_dict
        ret = []
        for key in inp:
            if  isinstance(inp[key], dict) :
                if len(inp[key]) != 0 :
                    ret.append(self.to_mongodb_filter(inp[key],
                                                previous_key=previous_key+'.'+key,
                                                ))
                else:
                    ret.append({previous_key+'.'+key:inp[key]})
            else:
                ret.append({previous_key+'.'+key:inp[key]})
        ret_p = []
        for item_1 in ret:
            if isinstance(item_1, list):
                for item_2 in item_1:
                    ret_p.append(item_2)
            else:
                ret_p.append(item_1)
        return ret_p
    
    @property
    def to_serial_dict(self):
        return serializer(self.mapping)
